"""
Tools for generating and manipulating LSD profiles.  Includes calculating
longitudinal magnetic fields and radial velocities.  Also includes a wrapper
around LSDpy.
"""

import copy
import warnings
import numpy as np
import matplotlib.pyplot as plt
import scipy.special as specialf
from scipy.optimize import curve_fit

from .obsSpec import Spectrum, read_spectrum

# use a version dependent name for trapezoidal integration,
# since Numpy 2.0 changed names (this could also be a try except)
if np.lib.NumpyVersion(np.__version__) >= '2.0.0':
    from numpy import trapezoid as _trapezoid
else:
    from numpy import trapz as _trapezoid

###################################
###################################
class LSD:
    """
    Holds the LSD profile data.

    This usually includes numpy arrays:

    * vel - velocity grid for the LSD profile
    * specI - the Stokes I profile
    * specSigI - the uncertainties on Stokes I
    * specV - the polarization profile, most often Stokes V
    * specSigI - the uncertainties on the polarization profile
    * specN1 - the Null1 profile
    * specSigN1 - the uncertainties on specN1
    * specN2 - the Null2 profile, if it exists, otherwise zeros
    * specSigN2 - the uncertainties on Null2 if they exist
    * header - an optional text header for the LSD profile file

    And integers:
      
    * numParam - The number of Stokes I, V, & Null profiles used (usually 1, 3, or 4)
    * npix - the number of pixels (points in velocity) in the LSD profile
    """
    def __init__(self, vel, specI, specSigI, specV=None, specSigV=None,
                 specN1=None, specSigN1=None, specN2=None, specSigN2=None,
                 header=None):
        """
        Initialize an LSD profile, using data from an existing profile. 

        :param vel: velocity grid for the LSD profile
        :param specI: the Stokes I profile
        :param specSigI: the uncertainties on Stokes I
        :param specV: the polarization profile, usually Stokes V
        :param specSigV: the uncertainties on the polarization profile 
        :param specN1: the Null1 profile
        :param specSigN1: the uncertainties on Null1
        :param specN2: the Null2 profile, if it exists, otherwise it is all 0s (optional)
        :param specSigN2: the uncertainties on Null2 if there are any,
                          otherwise all 0s (optional)
        :param header: the header of the LSD profile file, if it exists (optional) 
        """
        self.vel = vel
        self.specI = specI
        self.specSigI = specSigI
        self.npix = vel.size
        self.numParam = 1
        
        if(specV is not None and specSigV is not None):
            self.specV = specV
            self.specSigV = specSigV
            self.numParam = 2
        else:
            self.specV = np.zeros(self.npix)
            self.specSigV = np.zeros(self.npix)
            
        if(specN1 is not None and specSigN1 is not None):
            self.specN1 = specN1
            self.specSigN1 = specSigN1
            self.numParam = 3
        else:
            self.specN1 = np.zeros(self.npix)
            self.specSigN1 = np.zeros(self.npix)
        
        if(specN2 is not None and specSigN2 is not None):
            self.specN2 = specN2
            self.specSigN2 = specSigN2
            self.numParam = 4
        else:
            self.specN2 = np.zeros(self.npix)
            self.specSigN2 = np.zeros(self.npix)
        
        if(header is not None):
            self.header = header
        else:
            self.header = ""
    
    def __len__(self):
        """
        Return the length of each array in an LSD profile.
        They should all be the same length, so it just returns the length of
        the velocity array. 

        :rtype: int
        """
        return len(self.vel)

    def __getitem__(self, key):
        """
        Returns an LSD with only the values at the specified index(s).

        :param key: the index or slice being checked

        :rtype: LSD
        """
        vel_s = self.vel[key]
        specI_s = self.specI[key]
        specSigI_s = self.specSigI[key]
        specV_s = self.specV[key]
        specSigV_s = self.specSigV[key]
        specN1_s = self.specN1[key]
        specSigN1_s = self.specSigN1[key]
        specN2_s = self.specN2[key]
        specSigN2_s = self.specSigN2[key]
        slice_prof = LSD(vel_s, specI_s, specSigI_s, specV_s, specSigV_s,
                         specN1_s, specSigN1_s, specN2_s, specSigN2_s, self.header)
        slice_prof.numParam = self.numParam
        return slice_prof

    def __setitem__(self, key, newval):
        """
        Sets all values of the LSD object at the specified location equal to the
        input profile's values.

        :param key: the index or slice being overwritten
        :param newval: LSD object used to replace the overwritten values
        """
        if not(isinstance(newval, LSD)):
            raise TypeError()
        else:
            self.vel[key] = newval.vel
            self.specI[key] = newval.specI
            self.specSigI[key] = newval.specSigI
            self.specV[key] = newval.specV
            self.specSigV[key] = newval.specSigV
            self.specN1[key] = newval.specN1
            self.specSigN1[key] = newval.specSigN1
            self.specN2[key] = newval.specN2
            self.specSigN2[key] = newval.specSigN2

    def save(self, fname):
        """
        Save the LSD profile to a file.
        This saves to a text file in Donati's format.
        
        :param fname: the name of the file the LSD profile to saves to.
        """
        
        oFile = open(fname, 'w')
        if self.header != None:
            if self.header == "":
                #For blank headers include some placeholder text
                oFile.write('#LSD profile\n')
            else:
                oFile.write(self.header)
                #Make sure there is a line break after the first header text
                if self.header[-1] != '\n': oFile.write('\n')
            if self.numParam <= 3:
                oFile.write(' {:d} 6\n'.format(self.npix))
            elif self.numParam == 4:  #if Null1 & Null2 exist
                oFile.write(' {:d} 8\n'.format(self.npix))

        if self.numParam <= 3:
            for i in range(self.npix):
                oFile.write('{:>12.6f} {:>13.6e} {:>13.6e} {:>13.6e} {:>13.6e} {:>13.6e} {:>13.6e}\n'.format(
                    self.vel[i], self.specI[i], self.specSigI[i], self.specV[i],
                    self.specSigV[i], self.specN1[i], self.specSigN1[i]))
        elif self.numParam == 4:  #if Null1 & Null2 exist
            for i in range(self.npix):
                oFile.write('{:>12.6f} {:>13.6e} {:>13.6e} {:>13.6e} {:>13.6e} {:>13.6e} {:>13.6e} {:>13.6e} {:>13.6e}\n'.format(
                    self.vel[i], self.specI[i], self.specSigI[i], self.specV[i],
                    self.specSigV[i], self.specN1[i], self.specSigN1[i],
                    self.specN2[i], self.specSigN2[i]))
        oFile.close()
        return

    def norm(self, normValue):
        """
        Return a renormalized LSD profile. Divides the I, V, and Null 
        profiles, and their uncertainties, by a value.
        
        :param normValue: the value to renormalize (divide) the LSD profile by
        :rtype: LSD
        """
        new = LSD(self.vel, 
                  self.specI/normValue, self.specSigI/normValue, 
                  self.specV/normValue, self.specSigV/normValue,
                  self.specN1/normValue, self.specSigN1/normValue, 
                  self.specN2/normValue, self.specSigN2/normValue, 
                  self.header)
        new.numParam = self.numParam
        return new

    def vshift(self, velShift):
        """
        Return an LSD profile, with a shift to the velocities of the LSD profile.
        
        :param velShift: the change in velocity to be added
        :rtype: LSD
        """
        new = LSD(self.vel-velShift, 
                  self.specI, self.specSigI, 
                  self.specV, self.specSigV,
                  self.specN1, self.specSigN1, 
                  self.specN2, self.specSigN2, 
                  self.header)
        new.numParam = self.numParam
        return new

    def scale(self, scale_int, scale_pol):
        """
        Return an LSD profile with rescaled amplitudes of the LSD profile
        (see also set_weights()).
        
        :param scale_int: scale the intensity profile by this
        :param scale_pol: scale the polarization and null profiles by this
        :rtype: LSD
        """

        new = LSD(self.vel, 
                  1.0 - ((1.0-self.specI)*scale_int), self.specSigI*scale_int, 
                  self.specV*scale_pol, self.specSigV*scale_pol,
                  self.specN1*scale_pol, self.specSigN1*scale_pol, 
                  self.specN2*scale_pol, self.specSigN2*scale_pol, 
                  self.header)
        new.numParam = self.numParam
        
        # If we have scaled the profile we may need to adjust 
        # the weights in the header. Parse the old header,
        # and if necessary re-write the portion with the weights.
        if 'normalizing: d=' in self.header:
            ihead = self.header.index('normalizing: d=') - 1
            depth = 0; lande = 0; wl0 = 0
            #Get the old depth
            indDepth = self.header.find('d=')
            if indDepth >= 0:
                depth = float(self.header[indDepth+2:].split()[0])
            #Get the old Lande factor
            indLande = self.header.find('lande=')
            if indLande >= 0:
                lande = float(self.header[indLande+6:].split()[0])
            #Get the old wavelength
            indWl0 = self.header.find('wl=')
            if indWl0 >= 0:
                wl0 = float(self.header[indWl0+3:].split()[0])

            depth *= scale_int
            wl0 *= scale_pol/scale_int
            sFormat = (' normalizing: d={:5.3f} lande={:5.3f} wl={:6.1f} '
                       '(I norm weight {:5.3f}, V norm weight {:7.3f})\n')
            headerAdd = sFormat.format(depth, lande, wl0, depth,
                                       depth*lande*wl0)
            headder = self.header[:ihead] + headerAdd
            new.header = headder

        return new

    def set_weights(self, wint_old, wpol_old, wint_new, wpol_new):
        """
        Return an LSD profile with different mask normalization/scaling weights
        (see also scale()).
        
        :param wint_old: The current intensity weight (d)
        :param wpol_old: The current polarization weight (g*d*lambda)
        :param wint_new: The new intensity weight (d)
        :param wpol_new: The new polarization weight (g*d*lambda)
        :rtype: LSD
        """
        return self.scale(wint_new/wint_old, wpol_new/wpol_old)

    def coadd(self,*args):
        """
        coadd this LSD profile with other LSD profiles

        Currently the LSD profiles must all be on the same velocity grid,
        interpolation is not supported.  Recommend extracting all LSD
        on the same velocity grid to avoid having to interpolate.
        Coadding here essentially averages LSD profiles weighted
        by 1/sigma**2. 

        :param args: other LSD objects (or a list or tuple of them)
                     to coadd with this one
        :rtype: LSD
        """
        #Check if the user passed a list or tuple of spectra (unwrap it)
        _args = args
        if len(args) > 0:
            if isinstance(args[0], list) or isinstance(args[0], tuple):
                if isinstance(args[0][0], LSD): _args = args[0]

        #Set up a new LSD profile for a weighted average, starting with self
        #weight by 1/sigma**2, and save the sum of the weights
        weightI = 1./self.specSigI**2
        weightV = weightN1 = weightN2 = np.zeros_like(self.vel)
        if self.numParam > 1:
            weightV = 1./self.specSigV**2
        if self.numParam > 2:
            weightN1 = 1./self.specSigN1**2
        if self.numParam > 3:
            weightN2 = 1./self.specSigN2**2
        prof = LSD(self.vel.copy(), self.specI*weightI, weightI,
                   self.specV*weightV, weightV,
                   self.specN1*weightN1, weightN1,
                   self.specN2*weightN2, weightN2, self.header)
        prof.numParam = self.numParam

        for arg in _args:
            #basic error checking
            if not isinstance(arg, LSD):
                raise TypeError('LSD.coadd can only use LSD objects!')
            if arg.numParam != self.numParam:
                raise ValueError('LSD.coadd requires LSD profiles with the '
                                 'same number of columns (Stokes parameters)!')
            if np.any(arg.vel != self.vel):
                raise ValueError('LSD.coadd currently requires all LSD profiles'
                                 ' to have the same velocity grid! '
                                 '(Recommend re-calculating all profiles on  '
                                 'the same velocity grid)')
                                 
            #Add this profile, weighted by 1/sigma**2
            weightI = 1./arg.specSigI**2
            prof.specI += arg.specI*weightI
            prof.specSigI += weightI
            if self.numParam > 1:
                weightV = 1./arg.specSigV**2
                prof.specV += arg.specV*weightV
                prof.specSigV += weightV
            if self.numParam > 2:
                weightN1 = 1./arg.specSigN1**2
                prof.specN1 += arg.specN1*weightN1
                prof.specSigN1 += weightN1
            if self.numParam > 3:
                weightN2 = 1./arg.specSigN2**2
                prof.specN2 += arg.specN2*weightN2
                prof.specSigN2 += weightN2
            
        #Once all profiles have been added, complete the weighted average
        prof.specI /= prof.specSigI
        prof.specSigI = np.sqrt(1./prof.specSigI)
        if self.numParam > 1:
            prof.specV /= prof.specSigV
            prof.specSigV = np.sqrt(1./prof.specSigV)
        if self.numParam > 2:
            prof.specN1 /= prof.specSigN1
            prof.specSigN1 = np.sqrt(1./prof.specSigN1)
        if self.numParam > 3:
            prof.specN2 /= prof.specSigN2
            prof.specSigN2 = np.sqrt(1./prof.specSigN2)
        return prof
    
    def plot(self, figsize=(10,10), sameYRange=True, plotZeroLevel=True,
             fig=None, **kwargs):
        """
        Plot the LSD profile

        This generates a matplotlib figure. To display the figure, after
        running this function, use 'import matplotlib.pyplot as plt' and
        'plt.show()'. To save the figure, you can use 
        'fig.savefig("fileName.pdf")' on the new figure object.
        
        :param figsize: the size of the figure being created
        :param sameYRange: optionally set the Stokes V and Null plots to have
                           the same y scale
        :param plotZeroLevel: optionally include a dashed line at 0 for
                              the V and Null plots
        :param fig: optionally, a matplotlib figure used for plotting the LSD 
                    profile. By default (or if None) a new figure will be
                    generated.
        :return: a matplotlib figure object, and an axes object,
                 containing the plot. 
        """

        #Chose y-axis limits for the V and N plots so that they can have 
        #the same scale.  Use 1.05 time biggest divergence from zero
        plotVLims = np.max([np.abs(1.05*np.min(self.specV)),
                            np.abs(1.05*np.max(self.specV)),
                            np.abs(1.05*np.min(self.specN1)),
                            np.abs(1.05*np.max(self.specN1)),
                            np.abs(1.05*np.min(self.specN2)),
                            np.abs(1.05*np.max(self.specN2)),
                            1.05*np.max(self.specSigV)])
        
        #Get the number of Stokes parameters to plot
        #(Check the number of non-zero data field)
        #Support just Stokes I; Stokes I, V, Null1; Stokes I, V, Null1, Null2
        npar = self.numParam

        firstPlt = True
        if fig == None:
            fig, ax = plt.subplots(npar, 1, figsize=figsize, sharex=True)
        else:
            ax = fig.axes
            firstPlt = False

        #set up a custom color cycle, with black then the default colors
        if firstPlt:
            prop_cycle = plt.rcParams['axes.prop_cycle']
            colorList = ['#000000'] + prop_cycle.by_key()['color']

        #If there is only 1 plot (only a Stokes I spectrum),
        #wrap the matplotlib axis in a list to make it iterable
        if not (isinstance(ax, np.ndarray) or isinstance(ax, list)): ax = [ax]
        #protect against the number of axes not matching self.numParam
        for i, axi in enumerate(ax):
            if firstPlt: axi.set_prop_cycle(color=colorList)
            if i == len(ax) - 1: #Stokes I (always the last/bottom plot)
                axi.errorbar(self.vel, self.specI, yerr=self.specSigI, 
                             xerr=None, fmt='o', ms=3, ecolor='0.5', **kwargs)
                axi.set_xlim(xmin=np.min(self.vel),xmax=np.max(self.vel))
                axi.set_ylabel('I')
                axi.set_xlabel('Velocity (km/s)')
            elif (i == 0) and (npar > 1): #Stokes V
                axi.errorbar(self.vel, self.specV, yerr=self.specSigV, 
                             xerr=None, fmt='o', ms=3, ecolor='0.5', **kwargs)
                if plotZeroLevel: axi.axhline(y=0.0, ls='--', c='k', alpha=0.4)
                if sameYRange: axi.set_ylim(ymin=-plotVLims,ymax=plotVLims)
                axi.set_ylabel('V')
            elif (i == 1) and (npar > 2): #Null1
                axi.errorbar(self.vel, self.specN1, yerr=self.specSigN1, 
                             xerr=None, fmt='o', ms=3, ecolor='0.5', **kwargs)
                if plotZeroLevel: axi.axhline(y=0.0, ls='--', c='k', alpha=0.4)
                if sameYRange: axi.set_ylim(ymin=-plotVLims,ymax=plotVLims)
                axi.set_ylabel('N1')
            elif (i == 2) and (npar > 3): #Null2
                axi.errorbar(self.vel, self.specN2, yerr=self.specSigN2, 
                             xerr=None, fmt='o', ms=3, ecolor='0.5', **kwargs)
                if plotZeroLevel: axi.axhline(y=0.0, ls='--', c='k', alpha=0.4)
                if sameYRange: axi.set_ylim(ymin=-plotVLims,ymax=plotVLims)
                axi.set_ylabel('N2')
        
        plt.subplots_adjust(hspace=.0)
        return fig, ax

    def fit_gaussian_rv(self, velrange=None, plotFit=False, fullOutput=False,
                        scaleErrs=False):
        """
        Fit a Gaussian function to this LSD profile, to determine a
        radial velocity.
        
        :param velrange: Range of velocities included in the fit
                         (a tuple or list with 2 elements).
                         If not given, the whole range will be used.
        :param plotFit: If True, plot the fit to the LSD profile with matplotlib
        :param fullOutput: If True, return all the Gaussian best fit parameters,
                           rather than just the RV and RV error
        :param scaleErrs: If True, scale the errors on the fitting parameters by
                          the square root of the reduced chi^2
        :return: Tuple of best fit RV and RV error;
                 If fullOutput is True, then also includes continuum & error,
                 amplitude & error, and width & error
        """

        if velrange == None:
            velrange = (self.vel[1], self.vel[-2])
        elif not(isinstance(velrange, list) or isinstance(velrange, tuple)):
            raise TypeError('velrange in fitGaussianRV should be a list or '
                            +'tuple, with two elements')
        indVelUse = (self.vel >= velrange[0]) & (self.vel <= velrange[1])

        #Set the initial estimates for the fitting parameters
        indMinI = np.argmin(self.specI[indVelUse])
        initVel = self.vel[indVelUse][indMinI]
        initCont = 1.0
        initAmpl = 0.1
        initWidth = 10.
        param0 = [initCont, initAmpl, initVel, initWidth]

        #run the fitting routine
        fitVal, covarVal = curve_fit(_gaussProf, self.vel[indVelUse],
                                     self.specI[indVelUse], p0=param0,
                                     sigma=self.specSigI[indVelUse],
                                     absolute_sigma=(not scaleErrs))
        fitSigma = np.sqrt(np.diag(covarVal))
        
        fitCont = fitVal[0]
        fitAmpl = fitVal[1]
        fitVel = fitVal[2]
        fitWidth = fitVal[3]
        fitContErr = fitSigma[0]
        fitAmplErr = fitSigma[1]
        fitVelErr = fitSigma[2]
        fitWidthErr = fitSigma[3]

        #Optionally plot the fit to the observed LSD profile
        if plotFit == True:
            plt.errorbar(self.vel, self.specI,  yerr=self.specSigI, c='grey')
            plt.errorbar(self.vel[indVelUse], self.specI[indVelUse],
                          yerr=self.specSigI[indVelUse], c='k')
            plt.plot(self.vel[indVelUse], _gaussProf(
                self.vel[indVelUse], fitCont, fitAmpl, fitVel, fitWidth), 'b.')
            plotSynVels = np.linspace(velrange[0], velrange[1], 1000)
            plt.plot(plotSynVels, _gaussProf(
                plotSynVels, fitCont, fitAmpl, fitVel, fitWidth), 'b')
            plt.xlabel('Velocity (km/s)')
            plt.ylabel('I')
            plt.show()

        #Optionally return all fit parameters
        if fullOutput == True:
            return (fitVel, fitVelErr, fitCont, fitContErr, fitAmpl, 
                    fitAmplErr, fitWidth, fitWidthErr)
        return fitVel, fitVelErr

    def cog_rv(self, velrange, norm='auto'):
        '''
        Calculate a radial velocity using the center of gravity
        (i.e. first moment) of the Stokes I profile.

        This uses function the cog_I function, but provides the convenience
        of estimating the continuum level automatically and performing
        the calculation on a user specified velocity range.

        :param norm: calculation method for the continuum. The choices are:
                    'auto': the median of I outside of velrange;
                    or float: a user defined value to use for Ic.
        :param velrange: range of velocity to use for the determination of the
                    line center and the continuum.
                    (The integration range for the first moment.)
                    This should be a starting and ending velocity as two
                    elements inside a list (or tuple).
        :return:  the velocity of the center of gravity, and its uncertainty
        '''
        #Error checking on velrange
        if not (isinstance(velrange, list) or isinstance(velrange, tuple)):
            raise TypeError('LSD.cog_rv: velrange must be a list (or tuple) '
                             'with two values!')
        if len(velrange) < 2:
            raise ValueError('LSD.cog_rv: velrange must be a list (or tuple) '
                             'with two values!')
        #Get the portion of the LSD profile inside (and outside) velrange
        inside = (self.vel >= velrange[0]) & (self.vel <= velrange[1])
        lsd_in = self[inside]
        lsd_out = self[np.logical_not(inside)]

        #Calculate and apply the continuum normalization
        if isinstance(norm, str):
            #if norm == 'auto': #If multiple normalization methods are added
            norm_val = np.median(lsd_out.specI)
        else:
            norm_val = float(norm)

        #Calculate the COG, and its error, using cog_I
        cog_val, cog_err = lsd_in.cog_I(norm_val, fullOutput=True)
        
        return cog_val, cog_err
        
    
    def cog_I(self, Ic=1.0, fullOutput=False):
        '''
        Helper function to return the center of gravity of Stokes I for
        the LSD profile, for a given continuum level.

        :param Ic: the continuum level to use the the COG calculation
                   (float, default=1.0)
        :param fullOutput: If True, return the error of the velocity
                            of the center of gravity
        :return: the velocity of the center of gravity, optionally also its error
        '''
        # Computes the velocity of the center of gravity
        numerator = _trapezoid(self.vel * (Ic - self.specI), x=self.vel)
        denominator = _trapezoid(Ic - self.specI, x=self.vel)
        rv = numerator/denominator
        
        if fullOutput == True:
            #Propagate errors through the trapezoidal integration, with uneven 
            #pixel spacing, using the same math as _integrate_bz()
            deltav_arr = self.vel[1:] - self.vel[:-1]
            err_scale = deltav_arr[:-1] + deltav_arr[1:]
            err_scale = 0.5*np.concatenate((deltav_arr[:1], err_scale, deltav_arr[-1:]))
            
            numeratorSig = np.sqrt(np.sum((self.vel * self.specSigI * err_scale)**2))
            denominatorSig = np.sqrt(np.sum((self.specSigI * err_scale)**2))
            #Note, this bit of linear error propagation, assuming independent
            #uncertainties isn't quite perfect, since the numerator and denominator
            #depend on the same specSigI.  But the numerator dominates the total
            #uncertainty, so treatment of the denominator isn't really important.
            rvSig = np.abs(rv) * np.sqrt((numeratorSig/numerator)**2
                                         + (denominatorSig/denominator)**2)
            return rv, rvSig
        return rv

    def cog_IV(self, Ic=1.0):
        '''
        Helper function to return the center of gravity of (Stokes I)*(Stoke V)
        for the LSD profile, for a given continuum level.

        :param Ic: the continnum level to use the the COG calculation
                   (float, default=1.0)
        :return: the velocity of the center of gravity
        '''
        numerator = _trapezoid(self.vel * np.abs(self.specV * (Ic-self.specI)),
                             x=self.vel)
        denominator = _trapezoid(np.abs(self.specV * (Ic-self.specI)),
                               x=self.vel)
        return numerator/denominator

    def cog_V(self):
        '''
        Helper function to return the center of gravity of the Stokes V
        LSD profile.
        
        :return: the velocity of the center of gravity
        '''
        numerator = _trapezoid(self.vel * np.abs(self.specV), x=self.vel)
        denominator = _trapezoid(np.abs(self.specV), x=self.vel)
        return(numerator/denominator)        

    def cog_min(self):
        '''
        Helper function to return the velocity of the minimum of the
        Stokes I LSD profile.
        
        :return: the velocity of the Stokes I minimum
        '''
        
        cog_min = self.vel[self.specI.argmin()]
        if cog_min.size > 1:
            cog_min = cog_min[0]
        return cog_min

    def calc_fap(self):
        '''Helper function that calculates the FAP for Stokes V, null1,
        and null2, for the LSD profile.

        The False Alarm Probability (FAP) is the probability that the observed
        data are consistent with the null hypothesis of no magnetic field.
        In this case, the null hypothesis is a flat line in Stokes V
        (an offset from 0 is allowed to account for continuum polarization).
        The probability is evaluated from chi^2.
        
        If you would like a specific range in velocity, simply slice the LSD
        object beforehand.  Note that the calcBz function also returns the
        FAP inside the spectral line, over the same velocity range as used
        in the Bz calculation. 

        :return: the values FAP V, FAP N1, FAP N2. 
        '''

        approxDOF = (self.npix-1.)

        #'fitting' the flat line (essentially an average weighted by 1/sigma^2)
        contV = np.sum(self.specV/self.specSigV**2) / np.sum(1./self.specSigV**2)
        chi2V = np.sum(((self.specV - contV)/self.specSigV)**2)
        probV = 1.0-specialf.gammainc(approxDOF/2., chi2V/2.)
        #repeat for the Null1 and Null2 profiles (if they exist)
        probN1 = 0.
        probN2 = 0.
        
        if self.numParam > 2:
            contN1 = np.sum(self.specN1/self.specSigN1**2) / np.sum(1./self.specSigN1**2)
            chi2N1 = np.sum(((self.specN1 - contN1)/self.specSigN1)**2)
            probN1 = 1.0-specialf.gammainc(approxDOF/2., chi2N1/2.)
        if self.numParam > 3:
            contN2 = np.sum(self.specN2/self.specSigN2**2) / np.sum(1./self.specSigN2**2)
            chi2N2 = np.sum(((self.specN2 - contN2)/self.specSigN2)**2)
            probN2 = 1.0-specialf.gammainc(approxDOF/2., chi2N2/2.)
        
        return(probV, probN1, probN2)

    def calc_bz(self, cog='I', norm='auto', lambda0=500., geff=1.2,
                velrange=None, bzwidth=None, plot=True, **kwargs):
        '''Calculate the Bz of an LSD profile
        
        :param cog: The value or calculation method for the center of gravity.
                    The choices are:
                    'I': center of gravity of Stokes I,
                    'V': center of gravity of Stokes V,
                    'IV': center of gravity of I*V,
                    'min': velocity of the minimum of Stokes I,
                    or a float: a user defined value in km/s.
        :param norm: calculation method for the continuum. The choices are:
                    'auto': the median of I outside of velrange (if defined),
                    or the full range (if velrange is not defined),
                    or float: a user defined value to use for Ic.
        :param lambda0: wavelength of the transition in nanometers (default=500).
                    For an LSD profile, this is the lambda value the LSD
                    profile shape was scaled with.
        :param geff: effective Lande factor of the transition.
                    For an LSD profile, this is the geff value the LSD profile
                    shape was scaled with.
        :param velrange: range of velocity to use for the determination of the
                    line center and the continuum. If bzwidth is not given,
                    this range will also be used for the integration range
                    in the Bz calculation. 
                    If not given, the whole LSD profile will be used. 
        :param bzwidth: distance from the line center (from the cog) used 
                    for the integration range in the Bz calculation. 
                    One element = same on each side of line center.
                    Two elements = left and right of line center.
                    Not given (default): instead use velrange for this.
        :param plot: whether or not a graph is generated and returned.
        :param kwargs: additional keyword arguments are passed to the
                   matplotlib plotting routines, if a plot is generated.
        :return: a dictionary with Bz (in G) and FAP calculations,
                 optionally also a matplotlib figure.
        '''
        
        # velrange is used to identify the position of the line,
        # for calculating the cog, and for calculating the position
        # of the continuum.
        # If velrange is not defined, it will use the whole range.
        # The range for calculating Bz itself is controlled by bzwidth below
        if velrange != None:
            inside = np.logical_and(self.vel>=velrange[0], self.vel<=velrange[1])
            lsd_in = self[inside]
            lsd_out = self[np.logical_not(inside)]
        else:
            lsd_in=copy.deepcopy(self)
        
        # Check if norm is a string.
        if isinstance(norm, str):
            print('using AUTO method for the normalization')
            if velrange != None:
                print('  using the median of the continuum outside of the line')
                norm_val = np.median(lsd_out.specI)
            else:
                print('  no range in velocity given, using the median '
                      +'of the whole specI to determine continuum')
                norm_val = np.median(lsd_in.specI)
        else:
            norm_val = copy.copy(norm)
            print('using given norm value')

        # Check if cog is a string.
        if isinstance(cog, str):
            # calculate the cog for the chosen method
            if cog == 'I':
                cog_val = lsd_in.cog_I(norm_val)
            elif cog == 'min':
                cog_val = lsd_in.cog_min()
            elif cog == 'IV':
                cog_val = lsd_in.cog_IV(norm_val)
            elif cog == 'V':
                cog_val = lsd_in.cog_V()
            else:
                raise ValueError('calcBz got unrecognized value for cog: '
                                 +'{:}'.format(cog))
        else:
            cog_val=copy.copy(cog)
        
        # Now we define the position of the line for the Bz calculation itself.
        # If the keyword bzwidth is defined, we use that range from the chosen cog.
        if bzwidth == None:
            # No bzwidth. using vrange if defined
            if velrange != None:
                print('no bzwidth defined, using velrange to calculate Bz')
                lsd_bz = copy.copy(lsd_in)
                # saving the range for plotting later.
                p_bzwidth = np.copy(velrange)
            else:
                print('no bzwidth nor velrange defined, using full range to calculate Bz')
                p_bzwidth = [self.vel.min(), self.vel.max()]
                lsd_bz = copy.copy(self)
        else:
            # Check whether it is a numpy array
            if isinstance(bzwidth, list) or isinstance(bzwidth, tuple):
                if len(bzwidth) == 1:
                    # keeping the actual bz calculation range for plotting later.
                    p_bzwidth = [cog_val-bzwidth, cog_val+bzwidth]
                    lsd_bz = self[ np.logical_and(self.vel >= p_bzwidth[0],
                                                  self.vel <= p_bzwidth[1]) ]
                elif len(bzwidth) == 2:
                    p_bzwidth = [cog_val-bzwidth[0], cog_val+bzwidth[1]]
                    lsd_bz = self[ np.logical_and(self.vel >= p_bzwidth[0],
                                                  self.vel <= p_bzwidth[1]) ]
                else:
                    print('bzwidth has too many elements (need one or two)')
                    raise ValueError('bzwidth has too many elements '
                                     +'{:} (need 1 or 2)'.format(len(bzwidth)))
            else:
                p_bzwidth = [cog_val-bzwidth, cog_val+bzwidth]
                lsd_bz = self[ np.logical_and(self.vel >= p_bzwidth[0],
                                              self.vel <= p_bzwidth[1]) ]

    
        # Dealing with potential order overlaps 
        # (an uneven velocity grid should now be treated correctly)
        deltav_array = lsd_bz.vel[1:]-lsd_bz.vel[:-1]
        if np.any(deltav_array < 0.0):
            warnings.warn("\n The velocity array is not monotonically increasing. "
                          "\n There might be an order overlap in the region of the "
                          "observation used. \n calc_bz will sort the LSD "
                          "profile in velocity order. \n Make sure this is what "
                          "you want -- merge orders before running calc_bz()!",
                          stacklevel=2)
            lsd_bz = lsd_bz._sortvel()
            deltav_array = lsd_bz.vel[1:]-lsd_bz.vel[:-1]

        # Actual calculation of the Bz:

        # Call the integration function for each V, Null1, Null2 parameter
        blv, blvSig = _integrate_bz(lsd_bz.vel, lsd_bz.specV, lsd_bz.specSigV, 
                    geff, lambda0, cog_val, lsd_bz.specI, lsd_bz.specSigI, norm_val)
        if self.numParam > 2:
            bln1, bln1Sig = _integrate_bz(lsd_bz.vel, lsd_bz.specN1, lsd_bz.specSigN1,
                    geff, lambda0, cog_val, lsd_bz.specI, lsd_bz.specSigI, norm_val)
        else: bln1, bln1Sig = (0.0, 0.0)
        if self.numParam > 3:
            bln2, bln2Sig = _integrate_bz(lsd_bz.vel, lsd_bz.specN2, lsd_bz.specSigN2,
                    geff, lambda0, cog_val, lsd_bz.specI, lsd_bz.specSigI, norm_val)
        else: bln2, bln2Sig = (0.0, 0.0)
        
        # Get the FAP in the same range as the one used for Bz
        FAP_V, FAP_N1, FAP_N2 = lsd_bz.calc_fap()
        
        result = {
                'V bz (G)': blv,
                'V bz sig (G)': blvSig,
                'V FAP': FAP_V,
                'N1 bz (G)': bln1,
                'N1 bz sig (G)': bln1Sig,
                'N1 FAP': FAP_N1,
                'N2 bz (G)': bln2,
                'N2 bz sig (G)': bln2Sig,
                'N2 FAP': FAP_N2,
                'norm_method':norm,
                'norm': norm_val,
                'cog_method':cog,
                'cog': cog_val,
                'int. range start': p_bzwidth[0],
                'int. range end': p_bzwidth[1]
                }

        if plot:
            fig  = _plot_bz_calc(self, lsd_in, lsd_bz, velrange,
                            p_bzwidth, norm_val, cog_val, cog, **kwargs)
            return result,fig
        else:
            return result

    def calc_bis(self, Ic=1.,  cog='I', biswidth=None, plotFit=False, fullOutput=False):
        """
        Calculate the bisector velocity span to this LSD profile for a given
        continuum level.  This follows the line bisector and velocity span 
        definitions by Gray (1982) and Queloz et al. (2001), respectively.
        An uncertainty on the velocity span is estimated from the standard
        deviation of points in the upper and lower portions of the bisector.
        
        :param Ic: The continuum level to use the the BIS calculation
                   (float, default=1.0)
        :param cog: The value, or calculation method for the center of gravity.
                    The choices are:
                    'I': center of gravity of I,
                    'V': center of gravity of V,
                    'IV': center of gravity of I*V,
                    or a float: a user defined value in km/s.
        :param biswidth: Distance from the line center for the BIS calculation.
                    One element = same on each side of line center.
                    Two elements, left and right of line center.
                    Not defined: using the entire profile.
        :param plotFit: If True, plot the bisector curve and the LSD profile
                        with matplotlib
        :param fullOutput: If True, the return velocity span, top and bottom
                           velocity values, and line bisector.
                           If False just return the velocity span.
        :return: The velocity span, and error (Float).
                 If fullOutput is True then also include average velocity at
                 the top & bottom part of the bisector (Float) and numpy arrays 
                 with the line bisector & associated intensity levels.
        """

        lsd=copy.deepcopy(self)

        # Check if this is an emission line profile
        emission = lsd._check_emission(Ic)

        # Check if cog is a string.
        if isinstance(cog, str):
            # calculate the cog for the chosen method
            if cog == 'I':
                cog_val = lsd.cog_I(Ic)
            elif cog == 'min':
                cog_val = lsd.cog_min()
            elif cog == 'IV':
                cog_val = lsd.cog_IV(Ic)
            elif cog == 'V':
                cog_val = lsd.cog_V()
            else:
                raise ValueError('calc_bis got unrecognized value for cog: '
                                 '{:}'.format(cog))
        else:
            cog_val=copy.copy(cog)
            cog = 'fixed \; val.'

        # Get the velocity range
        if biswidth == None:
            print('no biswidth, using full velocity range to calculate BIS')
            velrange = (lsd.vel[1], lsd.vel[-2])
        else:
            # Check whether it is a numpy array
            if isinstance(biswidth, list) or isinstance(biswidth, tuple):
                if len(biswidth) == 1:
                    # keeping the actual bis calculation range for plotting later.
                    velrange = (cog_val-biswidth[0], cog_val+biswidth[0])
                elif len(bzwidth) == 2:
                    velrange = (cog_val-biswidth[0], cog_val+biswidth[1])
                else:
                    print('biswidth has too many elements (need one or two)')
                    raise ValueError('biswidth has too many elements '
                                     '{:} (need 1 or 2)'.format(len(biswidth)))
            else:
                velrange = (cog_val-biswidth, cog_val+biswidth)

        # Set the velocity window
        indVelUse = (lsd.vel >= velrange[0]) & (lsd.vel <= velrange[1])

        # Set the intensity levels for the line bisector calculation
        # split red and blue shifted halves of the line
        # also exclude a pixel very close to line center (if it exists)
        remove_center_vel = 0.5*np.mean(np.abs(lsd.vel[:-1] - lsd.vel[1:]))
        if not emission:
            _blu_use = ((lsd.vel < cog_val - remove_center_vel)
                        & (lsd.specI < Ic) & indVelUse)
            _red_use = ((lsd.vel > cog_val + remove_center_vel)
                        & (lsd.specI < Ic) & indVelUse)
            # get the portion of the extended red side of the line where 
            # I is monotonically increasing (for interpolating between later)
            _red_extra = (lsd.vel > cog_val) & (lsd.vel <= velrange[1]*2.)
            ind_red_extra = np.nonzero(_red_extra)[0]
            ind = 0
            for ind in range(ind_red_extra[1], ind_red_extra[-2]):
                if lsd.specI[ind] >= lsd.specI[ind+1]: break
            _red_extra[ind+1:] = False
        else:
            _blu_use = ((lsd.vel < cog_val - remove_center_vel)
                        & (lsd.specI > Ic) & indVelUse)
            _red_use = ((lsd.vel > cog_val + remove_center_vel)
                        & (lsd.specI > Ic) & indVelUse)
            # get the portion of the extended red side of the line where 
            # I is monotonically decreasing (for interpolating between later)
            _red_extra = (lsd.vel > cog_val) & (lsd.vel <= velrange[1]*2.)
            ind_red_extra = np.nonzero(_red_extra)[0]
            ind = 0
            for ind in range(ind_red_extra[1], ind_red_extra[-2]):
                if lsd.specI[ind] <= lsd.specI[ind+1]: break
            _red_extra[ind+1:] = False

        #Error trapping, in case too much of the line is above Ic
        if (np.count_nonzero(_red_use) < 1 or np.count_nonzero(_blu_use) < 1
            or np.count_nonzero(_red_extra) < 1):
            warnings.warn('Warning in calc_bis: using line that does not exist, '
                          'or is too shallow relative to input continuum level!\n'
                          'returning zeros!', stacklevel=2)
            if fullOutput == True:
                return 0.0, np.inf, 0.0, 0.0, np.array([0.0]), np.array([1.0])
            else:
                return 0.0, np.inf
        if np.count_nonzero(_red_extra) < np.count_nonzero(_red_use):
            warnings.warn('Warning in calc_bis: distortions in the shape of '
                          'the red side of the line may cause problems with '
                          'interpolation for the bisector.\n'
                          'Adjust biswidth or proceed with caution!',
                          stacklevel=2)
            _red_extra = _red_use

        # Define the line intensity limits
        specImax = np.max(lsd.specI[indVelUse])
        specImin = np.min(lsd.specI[indVelUse])

        # Get the maximum intensity variation in the LSD profile
        #deltaI = specImax - specImin
        if not emission:
            deltaI = Ic - specImin
        else:
            deltaI = specImax - Ic

        # Get intensity levels from blueshifted region
        specI_levels = lsd.specI[_blu_use]

        # Re-sample the velocity vector to match the selected Stokes I levels
        # Warning: Here, the velocity is written as a function of Stokes I 
        # to interpolate the velocity array easily. Because of that, it is 
        # necessary to reverse a couple of arrays to have an increasing
        # function of vel(specI).
        if not emission:
            specI_levels = specI_levels[::-1]
            rvel_levels = np.interp(specI_levels, lsd.specI[_red_extra],
                                    (lsd.vel - cog_val)[_red_extra])
            bvel_levels = (lsd.vel - cog_val)[_blu_use][::-1]
            # Define mask to exclude the top %5 of the line profile
            _ideep =  specI_levels <= (Ic - 0.05*deltaI)
        else:
            rvel_levels = np.interp(specI_levels, lsd.specI[_red_extra][::-1],
                                    (lsd.vel - cog_val)[_red_extra][::-1])
            bvel_levels = (lsd.vel - cog_val)[_blu_use]
            # Define mask to exclude the bottom %5 of the line profile
            _ideep =  specI_levels >= (Ic + 0.05*deltaI)
        # Bisector: half of the difference in velocity for a given intensity level
        bisector = (rvel_levels + bvel_levels)/2.0

        # Mask out points too close to the continuum level
        specI_levels = specI_levels[_ideep]
        bisector = bisector[_ideep]

        # Get average velocity at the top and bottom part of the bisector
        # the top part includes all points within 10–40 per cent of the full 
        # line depth and the bottom one those within 60–90 per cent.
        if not emission:
            _itop = ((specI_levels <= (Ic - 0.10*deltaI))
                     & (specI_levels >= (Ic - 0.40*deltaI)))
            _ibot = ((specI_levels <= (Ic - 0.60*deltaI))
                     & (specI_levels >= (Ic - 0.90*deltaI)))
        else:
            _ibot = ((specI_levels >= (Ic + 0.10*deltaI))
                     & (specI_levels <= (Ic + 0.40*deltaI)))
            _itop = ((specI_levels >= (Ic + 0.60*deltaI))
                     & (specI_levels <= (Ic + 0.90*deltaI)))
        # some error trapping in case there are no points in a depth range
        if np.count_nonzero(_itop) < 1:
            warnings.warn('Warning in calc_bis: not enough points in the '
                          'top part of the line to compute bisector span!'
                          '\n returning nan!', stacklevel=2)
            vt = np.nan
            vt_std = np.nan
        else:
            vt = np.nanmean(bisector[_itop])
            vt_std = np.nanstd(bisector[_itop])
        if np.count_nonzero(_ibot) < 1:
            warnings.warn('Warning in calc_bis: not enough points in the '
                          'bottom part of the line to compute bisector span!'
                          '\n returning nan!', stacklevel=2)
            vb = np.nan
            vb_std = np.nan
        else:
            vb = np.nanmean(bisector[_ibot])
            vb_std = np.nanstd(bisector[_ibot])

        # Get the bisector velocity span as defined in Queloz et al. 2001
        vs = vt - vb

        vs_std = np.sqrt(vt_std**2 + vb_std**2)

        # Optionally plot the bisector and the observed profile
        if plotFit == True:
            plt.figure(figsize=(5,5))
            plt.errorbar(lsd.vel, lsd.specI, yerr=lsd.specSigI,
                         c='k', zorder=1)
            scale_bisector = 20. # arbitrary scale for plotting
            plt.plot(bisector*scale_bisector, specI_levels, 'r.-', 
                     label=r'BIS$\times {:.0f}$'.format(scale_bisector),
                     lw=0.5, zorder=2)
            plt.axvline(cog_val, label=r'cog$_\mathrm{{{:}}}$'.format(cog))
            plt.axvline(velrange[0], c='k', ls=':', label='biswidth')
            plt.axvline(velrange[1], c='k', ls=':')
            plt.hlines(specI_levels, xmin=velrange[0], xmax=velrange[1],
                       alpha=0.33, color='c', zorder=0)
            if not emission:
                plt.hlines([Ic-0.10*deltaI, Ic-0.40*deltaI, Ic-0.60*deltaI,
                            Ic-0.90*deltaI], xmin=velrange[0], xmax=velrange[1],
                           alpha=0.33, color='r', ls=':', zorder=0)
            else:
                plt.hlines([Ic+0.10*deltaI, Ic+0.40*deltaI, Ic+0.60*deltaI,
                            Ic+0.90*deltaI], xmin=velrange[0], xmax=velrange[1],
                           alpha=0.33, color='r', ls=':', zorder=0)
            plt.xlabel('Velocity (km/s)')
            plt.ylabel('I')
            plt.legend(loc='lower left')
            plt.show()

        #Optionally return all fit parameters
        if fullOutput == True:
            return vs, vs_std, vt, vb, bisector, specI_levels
        else:
            return vs, vs_std
            
    def _sortvel(self):
        n = np.argsort(self.vel)
        return self[n]
    
    def _check_emission(self, Ic):
        '''
        Check if this is an emission line profile
        '''
        if np.nanmax(self.specI-Ic) > np.nanmax(Ic-self.specI):
            print("Assuming an emission line profile")
            return True
        else:
            return False
##################
##################

def _gaussProf(x, cont, ampl, center, width):
    """
    Define a simple Gaussian line profile (for fitting with)
    """
    y = cont - ampl*np.exp(-(x-center)**2/(2.*width**2))
    return y


def _integrate_bz(vel, spec, specSig, geff, lambda0, cog_val,
                  specI, specSigI, norm_val):
    '''
    Helper function that is used in the Bz calculations to integrate one Stokes parameter. 
    Internal.

    :param vel: (numpy array) the velocity array
    :param spec: (numpy array) the associated Stokes parameter array
    :param specSig: (numpy array) the error bar on the Stokes parameter
    :param geff: the effective lande factor of the profile
    :param lambda0: the nominal wavelength of the profile
    :cog_val: the chose center of gravity value (km/s)
    :specI: (numpy array) the associated Stokes I parameter
    :specSigI: (numpy array) the associated error bar on Stokes I
    :return: Bz and its error
    '''
    #Evaluate the integral equation for Bz.  Only intended for use in calcBz.
    
    # This is the constant for the Zeeman splitting
    # Lambda_B = constant * lambda0**2 B
    # lambda_B_constant = e/(4*pi*m_e*c) = 4.668644778304102e-05 in cgs units
    lambda_B_constant = 4.668644778304102e-12 #to match/cancel lambda0 in nm
    cvel = 2.99792458e5 #c in km/s to match/cancel the LSD profile velocities
    
    # set the velocity step for error propagation
    # (There might be some issues with order overlap if the function is used
    # with an LSD object that was calculated with individual_line function. )
    ###
    # For trapezoidal integration
    # A = sum of  0.5 [ x_(i+1) - x_(i) ]*[ Y_(i) + Y_(i+1) ] (from i=1 to i=N-1)
    # If we define dx_i = x_(i+1) - x_(i) we can rewrite this as:
    # A = sum of 0.5*dx_(i)*[ Y_(i) + Y_(i+1) ]  (from i=1 to i=N-1)
    # if we expand the sum, and gather all of the Y_i together, we get:
    # dx_1 Y_1 + [dx_1+dx_2]Y_2 + [dx_2+dx_3]Y_3 +
    #      ... + [dx_{N-2}+dx_{N-1}]Y_{N-1} + dx_{N-1} Y_N
    # Now, assuming uncertainties only on the Y_i, of s_i, the propagated error is
    # error^2 = 0.5 * ( (dx_1 s_1)**2 + ([dx_1+dx_2]s_2)**2 + ([dx_2+dx_3]s_3)**2 +
    #           ... +  ([dx_{N-2}+dx_{N-1}]*s_{N-1})**2 + (dx_{N-1} s_N)**2 )
    # numerically, if we have: vel = [x_1, x_2, x_3, ..., x_n],
    #              then: vel[1:]-vel[:-1] = [dx_1, dx_2, dx_3, ..., dx_{n-1}]
    # adding a zero at each end: [0, dx1, dx2, dx3, ..., dx_(n-1), 0], and doing [1:]+[:-1] 
    # will give us the array that we need to multiply with the s_i array. Et voila!
    ###
    #deltav_arr = vel[1:] - vel[:-1]
    #dx = np.hstack([0., deltav_arr, 0.])
    #err_scale = 0.5*(dx[:-1] + dx[1:])
    # or more intuitively to Colin (and marginally faster):
    deltav_arr = vel[1:] - vel[:-1]
    err_scale = deltav_arr[:-1] + deltav_arr[1:]
    err_scale = 0.5*np.concatenate((deltav_arr[:1], err_scale, deltav_arr[-1:]))

    # Calculation of the integral in the numerator of the Bz function
    # with a trapezoidal numerical integral
    fnum = _trapezoid((vel - cog_val) * spec, x=vel-cog_val) #in (km/s)^2
    sfnum = np.sqrt(np.sum(((vel - cog_val) * specSig * err_scale)**2))
    # for a constant pixel spacing deltav (with the end points treated exactly)
    #sfnumConst = np.sqrt((0.5*(vel[0] - cog_val)*specSig[0]*deltav)**2
    #               + np.sum(((vel[1:-1] - cog_val)*specSig[1:-1])**2)*deltav**2
    #               + (0.5*(vel[-1] - cog_val)*specSig[-1]*deltav)**2)
    
    # Calculation of the integral in the denominator of the Bz function
    # with a trapezoidal numerical integral
    # For the square error calculation, we propagate like we would for
    # summation in a numerical integral.
    ri0v = _trapezoid(norm_val - specI, x=vel) # in km/s
    si0v = np.sqrt(np.sum((specSigI * err_scale)**2)) # in km/s
    # for a constant pixel spacing deltav (with the end points treated exactly)
    #si0vConst = np.sqrt((0.5*specSigI[0]*deltav)**2
    #                    + np.sum(specSigI[1:-1]**2)*deltav**2
    #                    + (0.5*specSigI[-1]*deltav)**2)
    
    # Make the actual Bz calculation.
    # for the units to work out, lambda0 should be in nm
    bl = -1*fnum / (ri0v*geff*lambda0*cvel*lambda_B_constant)
    blSig = np.abs(bl * np.sqrt((sfnum/fnum)**2 + (si0v/ri0v)**2))
    return bl, blSig


def _plot_bz_calc(lsd, lsd_in, lsd_bz, velrange, p_bzwidth, norm_val, cog_val, cog,
                  **kwargs):
    """
    Generate a plot showing the center of gravity and integration ranges used
    in the calculation of Bz from an LSD profile.  Called by the calc_bz
    function. Mostly just for internal use.

    :param lsd: the full LSD profile to plot
    :param lsd_in: slice of an LSD profile with the range used for COG calculation
    :param lsd_bz: slice of an LSD profile with the range used for integration of Bz
    :param velrange: the velocity range used for COG calculation
    :param p_bzwidth: the velocity range used for integration of Bz
    :param norm_val: the continuum level used for normalization
    :param cog_val: the final COG used for Bz calculation
    :param cog: the input COG flag/value given by the user
    :return: a matplotlib figure object
    """
    #This function relies on the plot method of the LSD profile class
    fig, ax = lsd.plot(sameYRange=False, **kwargs)
    
    for item in ax:
        if velrange != None:
            item.axvline(x=velrange[0], ls='--', label='velrange')
            item.axvline(x=velrange[1], ls='--')
        item.axvline(x=p_bzwidth[0], ls='dotted', label='bzwidth')
        item.axvline(x=p_bzwidth[1], ls='dotted')
        
    ax[-1].axhline(y=norm_val, ls='--', c='pink', label='norm')
    
    # for the plot, calculate and display all of the possible methods
    # for calculating the cog.
    for item in ax:
        item.axvline(x=lsd_in.cog_min(), label='cog min I', lw=3, alpha=0.5, c='blue')
        item.axvline(x=lsd_in.cog_I(norm_val), label='cog I',lw=3, alpha=0.5, c='red')
        item.axvline(x=lsd_in.cog_IV(norm_val), label='cog I*V',lw=3, alpha=0.5, c='orange')
        item.axvline(x=lsd_in.cog_V(), label='cog V',lw=3, alpha=0.5, c='green')
        item.axvline(x=cog_val, label='chosen cog: {}'.format(cog), ls='--', c='k')
       
    ax[-1].legend(loc=0)
    
    red = lsd_bz.vel > cog_val
    blue = lsd_bz.vel < cog_val
    ax[0].fill_between(lsd_bz.vel[red], lsd_bz.specV[red], step='mid', color='red')
    ax[0].fill_between(lsd_bz.vel[blue], lsd_bz.specV[blue], step='mid', color='blue')
    if(len(ax) > 2):
        ax[1].fill_between(lsd_bz.vel[red], lsd_bz.specN1[red], step='mid', color='red')
        ax[1].fill_between(lsd_bz.vel[blue], lsd_bz.specN1[blue], step='mid', color='blue')
    if(len(ax) > 3):
        ax[2].fill_between(lsd_bz.vel[red], lsd_bz.specN2[red], step='mid', color='red')
        ax[2].fill_between(lsd_bz.vel[blue], lsd_bz.specN2[blue], step='mid', color='blue')
    
    return fig


def read_lsd(fname):
    """
    Read in an LSD profile and return an instance of the LSD class.
    
    The LSD profiles are in Donati's text format;
    however, the two lines of header in Donati's format are optional.
    
    :param fname: the name of the file containing the LSD profile
    :rtype: LSD
    """
    #check if this file has a header
    fcheck = open(fname, 'r')
    head1txt = fcheck.readline()
    head2txt = fcheck.readline()
    head3txt = fcheck.readline()
    fcheck.close()
    head1 = head1txt.split()
    head2 = head2txt.split()
    nskip = 2
    try: #Test if the first two lines look like LSD profile numbers
        float(head1[0])
        if(len(head2) > 2 and len(head1) == len(head2)):
            nskip = 0
    except(ValueError):
        nskip = 2

    header = None
    if(nskip > 0): header = head1txt

    #Check the number of columns in the LSD profile
    ncols = len(head3txt.split())
        
    #Read the profile, skipping the header
    __prof = np.loadtxt(fname, skiprows=nskip, unpack=True)
    
    if ncols == 9:
        vel = __prof[0,:]
        specI = __prof[1,:]
        specSigI = __prof[2,:]
        specV = __prof[3,:]
        specSigV = __prof[4,:]
        specN1 = __prof[5,:]
        specSigN1 = __prof[6,:]
        specN2 = __prof[7,:]
        specSigN2 = __prof[8,:]
        #Check for profiles with placeholder columns of 0 (LSDpy may do this!)
        if (np.all(specN2 == 0.0) and np.all(specSigN2 == 0.0)):
            ncols = 7
        elif (np.all(specV == 0.0) and np.all(specSigV == 0.0)
              and np.all(specN1 == 0.0) and np.all(specSigN1 == 0.0)):
            ncols = 3
        else:
            prof = LSD(vel, specI, specSigI, specV, specSigV,
                            specN1, specSigN1, specN2, specSigN2, header=header)
    if ncols == 7:
        vel = __prof[0,:]
        specI = __prof[1,:]
        specSigI = __prof[2,:]
        specV = __prof[3,:]
        specSigV = __prof[4,:]
        specN1 = __prof[5,:]
        specSigN1 = __prof[6,:]
        #Check for profiles with placeholder columns of 0 (LSDpy may do this!)
        if (np.all(specV == 0.0) and np.all(specSigV == 0.0)
              and np.all(specN1 == 0.0) and np.all(specSigN1 == 0.0)):
            ncols = 3
        else:
            prof = LSD(vel, specI, specSigI, specV, specSigV,
                        specN1, specSigN1, header=header)
    
    if ncols == 3:
        vel = __prof[0,:]
        specI = __prof[1,:]
        specSigI = __prof[2,:]
        prof = LSD(vel, specI, specSigI, header=header)

    #For unsupported formats or numbers of columns
    if ncols != 3 and ncols != 7 and ncols != 9:
        raise ValueError("Read an unexpected number of columns from "
                         +"{:}, can't read as an LSD profile.".format(fname))
    return prof

###################################

def run_lsdpy(obs, mask, outLSDName=None,
              velStart=-200.0, velEnd=+200.0, velPixel=None, 
              normDepth=0.2, normLande=1.2, normWave=500.0,
              removeContPol=True, trimMask=True, sigmaClipIter=0,
              sigmaClip=500, outModelName=None,
              plotLSD=True,  outPlotLSDName=None):
    """
    Run the LSDpy code and return an LSD object  
    (a convenience wrapper around the lsdpy.main() function).

    This requires LSDpy to be installed and in your Python path,
    which should be installed if you installed specpolFlow with pip
    (see https://github.com/folsomcp/LSDpy ).
    
    While some reasonable default values are included, pay attention to
    the parameters: velPixel, normDepth, normLande, normWave, and outName.
    
    Arguments are: 
    
    :param obs:           name of the input observed spectrum file
    :param mask:          name of the input LSD mask file
    :param outLSDName:    if provided, save the output LSD profile to this
                          file (Default = None, not saved)
    :param velStart:      float, starting velocity for the LSD profile (km/s)
    :param velEnd:        float, ending  velocity (km/s)
    :param velPixel:      float, velocity pixel size (km/s). If not provided,
                          this will be the median pixel size in the observation,
                          however it is recommended to set this explicitly.
    :param normDepth:     float, normalizing line depth
    :param normLande:     float, normalizing effective Lande factor
    :param normWave:      float, normalizing wavelength
    :param removeContPol: flag for whether continuum polarization is 
                          subtracted from the LSD profile
                          (Default = True)
    :param trimMask:      flag for whether very closely spaced lines 
                          should be removed from the line mask
                          (Default = True)
    :param sigmaClipIter: int, number of iterations for sigma clipping, 
                          rejecting possible bad pixels based on the fit to
                          Stokes I. Set to 0 for no sigma clipping.
                          (Default = 0, no sigma clipping)
    :param sigmaClip:     float, if sigma clipping, reject pixels where the
                          observation differs from the model by more than this
                          number of sigma.  Should be a large value so only very
                          bad pixels are rejected.
                          (Default = 500.)
    :param outModelName:  if provided, save the output model spectrum to this
                          file  (Default = None, not saved)
    :param plotLSD:       flag for whether to plot the LSD profile
                          (using matplotlib) (Default = True)
    :param outPlotLSDName: if provided, save the the plotted figure of the LSD 
                           profile  to this file (e.g. 'figProf.pdf'). Supports
                           the output file formats using the matplotlib savefig
                           function.
    :return: an LSD profile object, and a 'Spectrum' object containing the model
                          spectrum (the fit to the observation,
                          i.e. the convolution of the line mask and LSD profile)
    """
    import LSDpy

    # Passing LSD calculation controls to the LSDpy.lsd call
    interpMode = 1
    if removeContPol is False:
        _removeContPol = 0
    else:
        _removeContPol = 1
    if trimMask is False:
        _trimMask = 0
    else:
        _trimMask = 1

    # Passing figure generation and saving controls to the LSDpy.lsd call
    if plotLSD is False:
        _fLSDPlotImg = 0
    else:
        _fLSDPlotImg = 1
    if outPlotLSDName is None:
        _fSavePlotImg = 0
        _outPlotImgName = ''
    else:
        _fSavePlotImg = 1
        _outPlotImgName = outPlotLSDName

    # Make a default calculation for the pixel spacing if not user-defined
    if velPixel is None:
        #Get average observed wavelength spacing
        cvel = 2.99792458e5 #c in km/s
        obsSp = read_spectrum(obs)
        wlStep = obsSp.wl[1:] - obsSp.wl[:-1]
        medianVelPix = np.median(wlStep/obsSp.wl[:-1]*cvel)
        #indWlSteps = ((wlStep > 0.) & (wlStep < 0.5))
        #meanVelPix = np.average(wlStep[indWlSteps]/obsSp.wl[:-1][indWlSteps]*cvel)        
        velPixel = round(medianVelPix, ndigits=2)

    # Set the controls to the call to LSDpy for saving the files if needed.
    if outModelName is None:
        _outModelName = ''
    else:
        _outModelName = outModelName
    _fSaveModelS = 1     # and set the model spectrum calculation flag

    if outLSDName is None:
        _outLSDName = ''
    else:
        _outLSDName = outLSDName
        
    # Make the call to LSDpy.py
    vel, sI, sIerr, sV, sVerr, sN1, sN1err, headerTxt, specList = LSDpy.lsd(
        obs, mask, _outLSDName, velStart, velEnd, velPixel, 
        normDepth, normLande, normWave, _removeContPol, _trimMask, 
        sigmaClipIter, sigmaClip, interpMode, _fSaveModelS, _outModelName, 
        _fLSDPlotImg, _fSavePlotImg, _outPlotImgName)

    prof = LSD(vel, sI, sIerr, sV, sVerr, sN1, sN1err, header=headerTxt)
    modelSpec = Spectrum(specList[0], specList[1], specList[2], specList[3],
                            np.zeros_like(specList[0]), np.zeros_like(specList[0]))
    
    return prof, modelSpec

###################################



#!/usr/bin/python3
#
# Calculate a longitudinal magnetic field (Bz) from an LSD profile.

import argparse
import matplotlib.pyplot as plt
try:
    import specpolFlow as pol
except ModuleNotFoundError:
    #If specpolFlow is not installed or in the python path
    #try guessing the location and adding it temporarily
    import sys, os
    loc0 = os.path.dirname(__file__) 
    sys.path.insert(0, os.path.join(loc0, '..', '..'))
    import specpolFlow as pol

def bz_cli():
    """Main function for running LSD.calc_bz() as a terminal program

    Takes no arguments, but uses command line parameters instead.
    """
    #Take input file names and velocity ranges as command line arguments,
    #with some additional optional control parameters.
    parser = argparse.ArgumentParser(description="""
    Calculate the longitudinal magnetic field Bz from LSD profiles.
    Prints Bz from Stokes V and null profiles.
    Also prints the detection 'false alarm probability' (FAP)
    (1 - detection probability) for the profile.""")
    parser.add_argument("fileList", nargs='*',
                        help='LSD profile file(s), to calculate Bz for.  '
                        +'Can be more than one file.')
    parser.add_argument("-v", "--velRange", nargs=2, type=float,
                        metavar=('VEL1', 'VEL2'), required=True, 
                        help='Starting and ending velocity for the range used '
                        +'to calculate line centre and integrate.')
    parser.add_argument("-g", "--Lande", type=float,
                        help='The effective Lande factor used to normalize '
                        +'(scale) the LSD profile.')
    parser.add_argument("-l", "--wavelength", type=float,
                        help='The wavelength, in nm, used to normalize '
                        +'(scale) the LSD profile.')
    parser.add_argument("-p", "--plotFit", action='store_true',
                        help='Optional, plot information about the line range '
                        +'and centre used.')
    args = parser.parse_args()
    #Process the command line parameters
    fileList = []
    for fileName in args.fileList:
        fileList += [fileName]
    velRange = args.velRange
    lande = args.Lande
    wl0 = args.wavelength
    plotFit = args.plotFit
    if fileList == []:
        print('Provide an LSD profile file!\n(try -h for more info)\n')

    results=[]
    #Run the Bz calculation on any files provided
    for fileName in fileList:
        lsd = pol.read_lsd(fileName)
        #If there isn't a Lande factor and wavelength provided,
        #see if there is a value in the LSD profile header
        if lande == None:
            indLande = lsd.header.find('lande=')
            if indLande >= 0:
                lande = float(lsd.header[indLande+6:].split()[0])
            else:
                print('A reference effective Lande factor is needed')
        if wl0 == None:
            indWl0 = lsd.header.find('wl=')
            if indWl0 >= 0:
                wl0 = float(lsd.header[indWl0+3:].split()[0])
            else:
                print('A reference effective Wavelength is needed\n')
        
        #The main Bz calculation
        res = lsd.calc_bz(cog='I', lambda0=wl0, geff=lande,
                          velrange=velRange, plot=plotFit)
        
        #If we want to plot this figure, first unpack the two values returned,
        #then we need to run the matplotlib show function.
        if plotFit:
            res, fig  = res
            plt.show()
        results += [res]

    #Print the results out
    nPar = lsd.numParam
    txtline = ('file                             Bz_V(G)  sigma_V  ratio FAP_V')
    if nPar >= 3: txtline += '      Bz_N1(G) sigma_N1  ratio FAP_N1'
    if nPar >= 4: txtline += '     Bz_N2(G) sigma_N2  ratio FAP_N2'
    print(txtline)
    for i, res in enumerate(results):
        txtline = '{:30s} {:+9.2f} {:8.2f} {:6.2f} {:9.3e}'.format(
                       fileList[i], res['V bz (G)'], res['V bz sig (G)'],
                       res['V bz (G)']/res['V bz sig (G)'], res['V FAP'])
        if nPar >= 3: txtline += ' {:+9.2f} {:8.2f} {:6.2f} {:9.3e}'.format(
                res['N1 bz (G)'], res['N1 bz sig (G)'],
                res['N1 bz (G)']/res['N1 bz sig (G)'], res['N1 FAP'])
        if nPar >= 4: txtline += ' {:+9.2f} {:8.2f} {:6.2f} {:9.3e}'.format(
                res['N2 bz (G)'], res['N2 bz sig (G)'],
                res['N2 bz (G)']/res['N2 bz sig (G)'], res['N2 FAP'])
        print(txtline)
        
    return

#For running this Python script as a terminal program
if __name__ == "__main__":
    bz_cli()



#!/usr/bin/python3
#
# Convert ESPaDOnS .fits files from the UPENA format to a text .s format.

import argparse
import matplotlib.pyplot as plt
try:
    import specpolFlow as pol
except ModuleNotFoundError:
    #If specpolFlow is not installed or in the python path
    #try guessing the location and adding it temporarily
    import sys, os
    loc0 = os.path.dirname(__file__) 
    sys.path.insert(0, os.path.join(loc0, '..', '..'))
    import specpolFlow as pol


def espadons_cli():
    """
    Main function for running the ESPaDOnS converter as a terminal program.
    
    Takes no arguments, but uses command line parameters instead.
    """
    import argparse
    parser = argparse.ArgumentParser(description='Convert FITS file spectra '
            'from the ESPaDOnS CADC archive format to text .s files. Output '
            'files as [filename]n.s for the pipeline normalized spectra, '
            '[filename]u.s for unnormalized spectra, and [filename].out for '
            'header information.')
    parser.add_argument("observation", nargs='*',
                        help='a list of FITS files to process.')
    args = parser.parse_args()

    flist=args.observation

    if flist == []:
        print('No files given')
    else:
        #Run the conversion funciton
        pol.converters.espadons(flist)
    return

################

#For running this scirpt as a terminal program
if __name__ == "__main__":
    espadons_cli()



"""
Tools for manipulating lists of spectral line data.
Includes tools for reading and parsing VALD version 3 line lists 
(typically 'long format', 'extract stellar' requests).
"""

import numpy as np

###################################

class LineList:
    """
    Container for a set of spectral line data, usually from VALD.

    This usually contains: 
     
    * nLines - number of lines in the line list
    * ion - list of species identifiers (element or molecule and ionization)
    * wl - array of wavelengths
    * loggf - array of oscillator strengths (log gf)
    * Elo - array of excitation potentials for the lower level in
      the transition (in eV)
    * Jlo - array of J quantum numbers for the lower level
    * Eu - array of excitation potentials for the upper level in
      the transition (in eV)
    * Jup - array of J quantum numbers for the upper level
    * landeLo - array of Lande factors for the lower level
    * landeUp - array of Lande factors for the upper level
    * landeEff - array of effective Lande factors for the transition
    * rad - array of radiative damping coefficients
    * stark - array of quadratic Stark damping coefficients
    * waals - array of van der Waals damping coefficients
    * depth - depth at the centre of the spectral line, as estimated by VALD
    * configLo - list of strings with the electron configuration and
      term symbols for the lower level
    * configUp - list of strings with the electron configuration and
      term symbols for the upper level
    * refs - list of references for the sources of the line data (optional)
    """
    def __init__(self, ion, wl, loggf, Elo, Jlo, Eup, Jup, landeLo, landeUp,
                 landeEff, rad, stark, waals, depth, configLo, configUp, refs):
        self.ion      = ion
        self.wl       = wl
        self.loggf    = loggf
        self.Elo      = Elo
        self.Jlo      = Jlo
        self.Eup      = Eup
        self.Jup      = Jup
        self.landeLo  = landeLo
        self.landeUp  = landeUp
        self.landeEff = landeEff
        self.rad      = rad
        self.stark    = stark
        self.waals    = waals
        self.depth    = depth
        self.configLo = configLo
        self.configUp = configUp
        self.refs     = refs
        self.nLines   = self.wl.size

    def __getitem__(self, key):
        """
        Returns a LineList with only the values at the specified index(s).

        :param key: the index or slice being checked
        :rtype: LineList
        """
        ion      = self.ion[key]
        wl       = self.wl[key]
        loggf    = self.loggf[key]
        Elo      = self.Elo[key]
        Jlo      = self.Jlo[key]
        Eup      = self.Eup[key]
        Jup      = self.Jup[key]
        landeLo  = self.landeLo[key]
        landeUp  = self.landeUp[key]
        landeEff = self.landeEff[key]
        rad      = self.rad[key]
        stark    = self.stark[key]
        waals    = self.waals[key]
        depth    = self.depth[key]
        configLo = self.configLo[key]
        configUp = self.configUp[key]
        refs     = self.refs[key]
        lList =  LineList(ion, wl, loggf, Elo, Jlo, Eup, Jup, landeLo,
                          landeUp, landeEff, rad, stark, waals, depth,
                          configLo, configUp, refs)
        return lList

    def __setitem__(self, key, newval):
        """
        Sets all values of the LineList at the specified location
        equal to the input LineList values.

        :param key: the index or slice being overwritten
        :param newval: LineList used to replace the values given by key
        """
        if not(isinstance(newval, LineList)):
            raise TypeError()
        else:
            self.ion[key]      = newval.ion
            self.wl[key]       = newval.wl
            self.loggf[key]    = newval.loggf
            self.Elo[key]      = newval.Elo
            self.Jlo[key]      = newval.Jlo
            self.Eup[key]      = newval.Eup
            self.Jup[key]      = newval.Jup
            self.landeLo[key]  = newval.landeLo
            self.landeUp[key]  = newval.landeUp
            self.landeEff[key] = newval.landeEff
            self.rad[key]      = newval.rad
            self.stark[key]    = newval.stark
            self.waals[key]    = newval.waals
            self.depth[key]    = newval.depth
            self.configLo[key] = newval.configLo
            self.configUp[key] = newval.configUp
            self.refs[key]     = newval.refs
            self.nLines   = self.wl.size

    def __len__(self):
        return self.nLines

    def __str__(self):
        """
        Generate a nicely formatted string of line data for printing
        """
        strList = []
        outStr = '\n'
        lines = self
        #check if this LineList contains arrays or just single values
        if isinstance(self.ion, str) and isinstance(self.wl, float):
            lines = [self]
        for line in lines:
            fmt = ("'{:s}',{:17.4f},{:7.3f},{:8.4f},{:5.1f},{:8.4f},{:5.1f},"
                   "{:7.3f},{:7.3f},{:7.3f},{:6.3f},{:6.3f},{:8.3f},{:6.3f},")
            if len(line.ion) == 3: fmt = fmt[:7] + ' ' + fmt[7:]
            strList.append(fmt.format(line.ion, line.wl, line.loggf,
                                line.Elo, line.Jlo, line.Eup, line.Jup,
                                line.landeLo, line.landeUp, line.landeEff,
                                line.rad, line.stark, line.waals, line.depth))
            strList.append("'  {:>88}'".format(line.configLo))
            strList.append("'  {:>88}'".format(line.configUp))
            strList.append("'{:}'".format(line.refs))
        return outStr.join(strList)

    def insert(self, i, newval):
        """
        Insert lines from a LineList at a specific index.  Returns a new 
        LineList combining the two line lists (does not operate in place)

        :param i: index to insert at
        :param newval: LineList of new values to insert
        :rtype: LineList
        """
        if not(isinstance(newval, LineList)):
            raise TypeError('LineList.insert: can only insert a LineList')

        #Re-allocating these arrays, particularly arrays of strings, is slow!
        ion      = np.insert(self.ion     , i, newval.ion     )
        wl       = np.insert(self.wl      , i, newval.wl      )
        loggf    = np.insert(self.loggf   , i, newval.loggf   )
        Elo      = np.insert(self.Elo     , i, newval.Elo     )
        Jlo      = np.insert(self.Jlo     , i, newval.Jlo     )
        Eup      = np.insert(self.Eup     , i, newval.Eup     )
        Jup      = np.insert(self.Jup     , i, newval.Jup     )
        landeLo  = np.insert(self.landeLo , i, newval.landeLo )
        landeUp  = np.insert(self.landeUp , i, newval.landeUp )
        landeEff = np.insert(self.landeEff, i, newval.landeEff)
        rad      = np.insert(self.rad     , i, newval.rad     )
        stark    = np.insert(self.stark   , i, newval.stark   )
        waals    = np.insert(self.waals   , i, newval.waals   )
        depth    = np.insert(self.depth   , i, newval.depth   )
        configLo = np.insert(self.configLo, i, newval.configLo)
        configUp = np.insert(self.configUp, i, newval.configUp)
        refs     = np.insert(self.refs    , i, newval.refs    )
        lList =  LineList(ion, wl, loggf, Elo, Jlo, Eup, Jup, landeLo,
                          landeUp, landeEff, rad, stark, waals, depth,
                          configLo, configUp, refs)
        return lList

    def save(self, fname):
        """
        Write a line list to a text file.
        This outputs using the VALD version 3 'extract stellar' 'long' format.
        
        A few details (e.g. references at the end of the file)
        are omitted since they are not saved in the LineList class.
        
        :param fname: the file name to save the output to
        """
        
        fOut = open(fname, 'w')
        fOut.write(("{:11.5f},{:12.5f},{:6d},{:7d},{:4.1f}, Wavelength "
                    "region, lines selected, lines processed, Vmicro\n"
                    ).format(self.wl[0], self.wl[-1], self.nLines, 999999, 0.))
        fOut.write("                                                 "
                   "                    Lande factors       "
                   "Damping parameters   Central\n")
        fOut.write("Spec Ion       WL_air(A)  log gf* E_low(eV) J lo "
                   "E_up(eV)  J up  lower   upper    mean   Rad.  "
                   "Stark   Waals   depth\n")
        fOut.write(str(self)+'\n')
        fOut.close()
        return
    
def line_list_zeros(nLines):
    """
    Generate a line list of zeros and blank text.
    
    Used by read_VALD
    (It can be a bit faster to allocate all the array space at once.)
    
    :param nLines: the number of lines in the LineList of zeros
    :rtype: LineList
    """
    ion      = np.tile(np.array([''], dtype='U6'), nLines)
    wl       = np.zeros(nLines)
    loggf    = np.zeros(nLines)
    Elo      = np.zeros(nLines)
    Jlo      = np.zeros(nLines)
    Eup      = np.zeros(nLines)
    Jup      = np.zeros(nLines)
    landeLo  = np.zeros(nLines)
    landeUp  = np.zeros(nLines)
    landeEff = np.zeros(nLines)
    rad      = np.zeros(nLines)
    stark    = np.zeros(nLines)
    waals    = np.zeros(nLines)
    depth    = np.zeros(nLines)
    configLo = np.tile(np.array([''], dtype='U128'), nLines)
    configUp = np.tile(np.array([''], dtype='U128'), nLines)
    refs = np.tile(np.array(['_          unknown source'],dtype='U180'), nLines)
    lList = LineList(ion, wl, loggf, Elo, Jlo, Eup, Jup, landeLo,
                     landeUp, landeEff, rad, stark, waals, depth,
                     configLo, configUp, refs)
    return lList

def read_VALD(fname):
    """
    Read a list of spectral line data from VALD and return a LineList.

    This expects VALD version 3 line list, in an 'extract stellar'
    'long' format.

    :param fname: the file name for the VALD line list.
    :rtype: LineList
    """
    fVald = open(fname, 'r')
    i = 0
    j = 0
    nLines = 0
    for txtLine in fVald:
        if i == 0:
            nLines = int(txtLine.split(',')[2])
            llist = line_list_zeros(nLines)
        #There should be 3 lines of header,
        #then 4 file lines for each set of spectra line data.
        if i > 2 and i < nLines*4+3:
            ii = (i-3)%4
            if ii == 0:
                #The line data are: Spec Ion, Wl(A), log gf, E_lower(eV), 
                #J_lower, E_upper(eV), J_upper, Lande factor lower, 
                #Lande factor upper, effective Lande factor, Rad. damping, 
                #Stark damping, van der Waals damping, central depth
                vals = txtLine.split(',')
                llist.ion[j]      = vals[0].strip('\'')
                llist.wl[j]       = float(vals[1])
                llist.loggf[j]    = float(vals[2])
                llist.Elo[j]      = float(vals[3])
                llist.Jlo[j]      = float(vals[4])
                llist.Eup[j]      = float(vals[5])
                llist.Jup[j]      = float(vals[6])
                llist.landeLo[j]  = float(vals[7])
                llist.landeUp[j]  = float(vals[8])
                llist.landeEff[j] = float(vals[9])
                llist.rad[j]      = float(vals[10])
                llist.stark[j]    = float(vals[11])
                llist.waals[j]    = float(vals[12])
                llist.depth[j]    = float(vals[13])
            if ii == 1:
                llist.configLo[j] = txtLine.strip(' \n\'')
            if ii == 2:
                llist.configUp[j] = txtLine.strip(' \n\'')
            if ii == 3:
                llist.refs[j] = txtLine.strip(' \n\'')
                j += 1
        i += 1
    
    fVald.close()
    return llist

###################################

def getEffectiveLande(landeLo, landeUp, Jlo, Jup):
    """
    Calculate an effective Lande factor for a transition.
    
    This uses the upper and lower level Lande factors and J quantum numbers.
    The effective Lande factor is similar to an average weighted by the
    splitting pattern of the line.  It parameterizes the displacement of
    the centre of gravity of the sigma components of a Zeeman split line.
    
    :param landeLo: Lower level Lande factor
    :param landeUp: Upper level Lande factor
    :param Jlo: Lower level J quantum number
    :param Jup: Upper level J quantum number
    :return: Effective Lande factor for the transition
    """

    #From, e.g., Degl'Innocenti & Landolfi 2004, eq. 3.44 
    #(originally from Shenstone and Blair (1929) I think)
    landeEff = 0.5*(landeUp + landeLo) + 0.25*(landeUp - landeLo)*(Jup*(Jup + 1) - Jlo*(Jlo + 1))
    
    return landeEff


def estimateLande(J, config, verbose=False):
    """
    Estimate a Lande factor for a level (if VALD doesn't have one).

    This supports levels in LS coupling, JJ coupling, and JK coupling.
    Levels in LK coupling are not supported.  This relies on VALD3's text 
    format for electron term symbols and configurations.
    
    :param J: The J quantum number for this level
    :param config: The text string from VALD with the electron configuration
                   and term symbol for the level
    :param verbose: Report warnings for levels in unsupported coupling schemes
                    if True
    :return: The Lande factor estimated for this level.  If the Lande factor
             can't be estimated this will be 99.0.
    """
    
    #Letters for L quantum numbers (note J is omitted in this notation)
    Lval = ('S','P','D','F','G','H','I','K','L','M','N','O','Q','R','T','U','V')
    lval = ('s','p','d','f','g','h','i','k','l','m','n','o','q','r','t','u','v')

    #In the rare case of a blank string for the electron configuration, abort
    if(len(config.split()) < 2): return 99.00
    
    coupling = config.split()[0]

    #For levels in LS coupling
    if coupling == 'LS':
        #Parse the text string
        L, S = get_LS_numbers(config)
        #and return 99 if that failed.
        if L == -1 or S == -1.0: return 99.00
        
        #The Lande factor in L-S coupling
        #(e.g. Degl'Innocenti & Landolfi 2004 eq. 3.8)
        #note, levels with J = 0 do not split at all (but break this equation)
        if J == 0.0:
            lande = 0.0
        else:
            lande = 1.0 + (J*(J + 1.) - L*(L + 1.) + S*(S + 1.)) \
                / (2.*J*(J + 1.))
        
    #For levels in JJ coupling
    #This covers 'J1-J2' 'J1-j' and (I think) 'j-j' cases
    #In this coupling scheme, individual electron l and s values couple
    #to make j  then the individual j couple to the total angular momentum J.
    #Also, subgroups of electrons can couple together to produce a combined
    #angular momentum J1 or J2, and those then combined to produce he total J.
    #
    #For this we want the J1, J2, L1 and L2 values, and possibly 
    #S1, S2 if they are given (if there are are subgroups of electrons).
    #This requires parsing an electron configuration as well as a term symbol.
    elif coupling == 'JJ':
        #Parse the text string
        L1, S1, J1, L2, S2, J2 = get_JJ_numbers(config)
        #and return 99 if that failed.
        if L1 < 0 or L2 < 0 or S1 < 0.0 or S2 < 0.0 or J1 < 0.0 or J2 < 0.0:
            return 99.0
        
        if J == 0.0:
            lande = 0.0
        else:
            #Landi Degl'Innocenti & Landolfi 2004 Polarization in Spectral Lines
            #Sect 3.1 (page 77, second equation, unnumbered after eq 3.8).
            #Also: W.C. Martin, R. Zalubas, and L. Hagan,
            #Atomic Energy Levels - The Rare-Earth Elements, 
            #Natl. Stand. Ref. Data Ser., Natl. Bur. Stand.
            #(U.S. Government Printing Office, Washington, 1978), No. 60.,
            #Sect 2.4
            gam1 = (J*(J+1.) + J1*(J1+1.) - J2*(J2+1.))/(2.*J*(J+1.))
            gam3 = (J*(J+1.) + J2*(J2+1.) - J1*(J1+1.))/(2.*J*(J+1.))
            if J1 == 0:
                gam2 = 0.
            else:
                gam2 = (J1*(J1+1.) + S1*(S1+1.) - L1*(L1+1.))/(2.*J1*(J1+1.))
            if J2 == 0: gam4 = 0.
            else:
                gam4 = (J2*(J2+1.) + S2*(S2+1.) - L2*(L2+1.))/(2.*J2*(J2+1.))
            lande = 1.0 + gam1*gam2 + gam3*gam4
        ##print('JJ lande {:6.3f}'.format(lande))
    
    #For levels in JK coupling
    elif coupling == 'JK':
        #Parse the text string
        L1, S1, J1, L2, S2, K = get_JK_numbers(config)
        #and return 99 if that failed.
        if L1 < 0 or L2 < 0 or S1 < 0.0 or S2 < 0.0 or J1 < 0.0 or K < 0.0:
            return 99.0
        
        #From W.C. Martin, R. Zalubas, and L. Hagan, 1978,
        #Atomic Energy Levels - The Rare-Earth Elements, 
        #Natl. Stand. Ref. Data Ser., Natl. Bur. Stand., Sect 2.4
        #(They cite: Wybourne, B. G., 1965, Spectroscopic properties of Rare
        #Earths, John Wiley & Sons, New York, N.Y., pg. 100),
        #which is for the more general J1-L2 case of JK coupling.
        #Landi Degl'Innocenti & Landolfi 2004 also provide a formula,
        #but only for the J1-l case of a subgroup coupling with one electron,
        #i.e. case when S2 = 1/2. We use the more general form here.
        if J == 0.0:
            lande = 0.0
        else:
            if J1 == 0.0:
                gam1 = 0.0
            else:
                gam1 = (J1*(J1+1.) - L1*(L1+1.) + S1*(S1+1.))/(2.*J1*(J1+1.))
            gam2 = (K*(K+1.) + J1*(J1+1.) - L2*(L2+1.))/((2.*J+1.)*(2.*K+1.))
            gam3 = (3.*J*(J+1.) - K*(K+1.) + S2*(S2+1.))/(2.*J*(J+1.))
            lande = 2.*gam1*gam2 + gam3
        ##print('JK lande {:6.3f}'.format(lande))
    
    else:
        #LK coupling is not supported, since I am not aware of a good, simple
        #approximation for calculating Lande factors in this coupling scheme.
        #Molecular lines are not supported here either.
        lande = 99.0
        if verbose:
            print('unsupported coupling scheme: ', config)
    return lande


def get_LS_numbers(config):
    """
    Extract the L and S quantum numbers from a VALD term symbol text string,
    for a level in LS coupling.
    
    :param config: the electron configuration and term symbol text string
    :return: The L and S quantum numbers (if a parsing error occurs,
             negative values are returned)
    """
    #Letters for L quantum numbers (note J is omitted in this notation)
    Lval = ('S','P','D','F','G','H','I','K','L','M','N','O','Q','R','T','U','V')
    
    #The last set of characters should be the term symbol in LS coupling.
    #We want the total orbital angular momentum quantum number L
    #and the total spin quantum number S.
    #This should be in the form (2S+1)L, where L is a capital letter
    #and (2S+1) is a number, possibly followed by * to indicate odd parity.
    #L value letters go like S=0, P=1, D=2, F=3, G=4, and so on
    L = -1
    S = -1.
    termSym = config.split()[-1]
    for i in range(len(termSym)):
        if termSym[i] in Lval:
            L = int(Lval.index(termSym[i]))
            #in case the multiplicity for S is 2 digits
            ii = i - 1
            if termSym[0:i].isnumeric(): ii = 0
            try:
                S = (float(termSym[ii:i]) - 1.)/2.
            except ValueError:
                print('Failed to parse S in LS coupling, from:\n', config)
                return -1, -1.
    if L < 0 or S < 0:
        print('Failed to parse L and S in LS coupling, from:\n', config)
        return -1, -1.
    return L, S


def get_JJ_numbers(config):
    """
    Extract the L1, S1, J1, L2, S2, and J2 quantum numbers for a level in JJ 
    coupling from VALD's electron configuration and term symbol text string.

    This should cover the J1-J2 case (where two subgroups combine to produce
    J1 and J2, which combine to produce the total J), the J1-j case (where
    one subgroup makes J1, which combines with an additional electron with j
    to produce J), and it should work for the j-j case (for two
    equivalent electrons with their own j, which combine to make J).
    
    :param config: the electron configuration and term symbol text string
    :return: The L1, S1, J1, L2, S2, and J2 quantum numbers
             (if a parsing error occurs, negative values are returned)
    """
    #Letters for L quantum numbers (note J is omitted in this notation)
    Lval = ('S','P','D','F','G','H','I','K','L','M','N','O','Q','R','T','U','V')
    lval = ('s','p','d','f','g','h','i','k','l','m','n','o','q','r','t','u','v')
    #The J1 and J2 values are in brackets () at the end as the term symbol
    termSym = config.split()[-1]
    econfig = config.split()[-2]
    if not ',' in termSym:
        print('Failed to parse J1, J2 in JJ coupling, from:\n', termSym)
        return -1, -1., -1., -1, -1., -1.
    sJ1, sJ2 = termSym.strip('*').strip('()').split(',')
    #Often J1 and J2 are fractions and need to be converted
    J1 = _parseFraction(sJ1, errMsg='Failed to parse J1 in JJ coupling:')
    J2 = _parseFraction(sJ2, errMsg='Failed to parse J2 in JJ coupling:')
    if np.isnan(J1) or np.isnan(J2): return -1, -1., -1., -1, -1., -1.
    #The L and possibly S values are given in the configuration.
    #If there is a subgroup of electrons there should be a capital letter
    #in (), giving the L value, and before that the number 2S+1.
    #If there is a single electron there should be a lower case letter
    #giving l (the last lower case letter in the string for l2) and s = 1/2.
    #The J_i values for a subgroup of electrons is given here in <>,
    #but we already have that from the term symbol so don't need to
    #extract it here.
    L1 = -1
    L2 = -1
    S1 = -1.
    S2 = -1.
    #The L2 (or l2) number should be the last letter in the string (upper or
    # lower case).  If the letter is a capital, try the proceeding number as
    #a multiplicity to get an S2 value, otherwise for a single electron
    #just use s2 = 1/2.
    for i2 in range(len(econfig)-1, 0, -1):
        #for a subgroup with L2
        if econfig[i2] in Lval:
            L2 = Lval.index(econfig[i2])
            ii = i2-1
            if econfig[i2-2:i2].isnumeric(): ii = i2-2
            try:
                S2 = (float(econfig[ii:i2])-1.)/2.
            except ValueError:
                print('Failed to parse S2 in JJ coupling, from:\n', config)
                return -1, -1., -1., -1, -1., -1.
            break
        #for a single e- with l2
        elif econfig[i2] in lval:
            L2 = lval.index(econfig[i2])
            S2 = 0.5
            break
    #Try to find L1, a capital (in brackets) before the L2 (or l2)
    for i1 in range(i2-1, 0, -1):
        if econfig[i1] in Lval:
            L1 = Lval.index(econfig[i1])
            ii = i1-1
            if econfig[i1-2:i1].isnumeric(): ii = i1-2
            try:
                S1 = (float(econfig[ii:i1])-1.)/2.
            except ValueError:
                print('Failed to parse S1 in JJ coupling, from:\n', config)
                return -1, -1., -1., -1, -1., -1.
            break
    #If there is no capital L1 for a subgroup, try to find l1 for a single e-
    #(the j-j case).
    if L1 == -1:
        for i1 in range(i2-1, 0, -1):
            if econfig[i1] in lval:
                L1 = lval.index(econfig[i1])
                S1 = 0.5
                break

    if L1 < 0 or L2 < 0 or S1 < 0.0 or S2 < 0.0:
        print('Failed to parse JJ coupling, from:\n', config)

    return L1, S1, J1, L2, S2, J2


def get_JK_numbers(config):
    """
    Extract the L1, S1, J1, L2, S2, and K quantum numbers for a level in JK
    coupling from VALD's electron configuration and term symbol text string.

    This is sometimes called J1-l or J1-L2 coupling.

    :param config: the electron configuration and term symbol text string
    :return: The L1, S1, J1, L2, S2, and K quantum numbers
             (if a parsing error occurs, negative values are returned)
    """
    #Letters for L quantum numbers (note J is omitted in this notation)
    Lval = ('S','P','D','F','G','H','I','K','L','M','N','O','Q','R','T','U','V')
    lval = ('s','p','d','f','g','h','i','k','l','m','n','o','q','r','t','u','v')
    
    termSym = config.split()[-1]
    econfig = config.split()[-2]

    K = -1.
    #The K value is in square brackets [] at the end as the term symbol
    if (not '[' in termSym) or (not ']' in termSym):
        print('Failed to parse K in JK coupling, from:\n', termSym)
        return -1, -1., -1., -1, -1., -1.
    sK = termSym.split('[')[1].split(']')[0]
    #K may be a fraction and need to be converted
    K = _parseFraction(sK, errMsg = 'Failed to parse K in JK coupling, from:')
    if np.isnan(K): return -1, -1., -1., -1, -1., -1.
    #The multiplicity (2*s+1) of the second electron or subgroup is usually
    #given just before the [], but since it could theoretically be omitted
    #we will infer it from the configuration below.
    
    L1 = -1
    L2 = -1
    S1 = -1.
    S2 = -1.
    J1 = -1.
    #The L2 (or l2) number should be the last letter in the string (upper or
    # lower case).  If the letter is a capital, try the proceeding number as
    #a multiplicity to get an S2 value, otherwise for a single electron
    #just use s2 = 1/2.
    for i2 in range(len(econfig)-1, 0, -1):
        #for a subgroup with L2
        if econfig[i2] in Lval:
            L2 = Lval.index(econfig[i2])
            ii = i2-1
            if econfig[i2-2:i2].isnumeric(): ii = i2-2
            try:
                S2 = (float(econfig[ii:i2])-1.)/2.
            except ValueError:
                print('Failed to parse S2 in JK coupling, from:\n', config)
                return -1, -1., -1., -1, -1., -1.
            break
        #for a single e- with l2
        elif econfig[i2] in lval:
            L2 = lval.index(econfig[i2])
            S2 = 0.5
            break
    #Try to find L1, a capital (in brackets) before the L2 (or l2)
    for i1 in range(i2-1, 0, -1):
        if econfig[i1] in Lval:
            L1 = Lval.index(econfig[i1])
            ii = i1-1
            if econfig[i1-2:i1].isnumeric(): ii = i1-2
            try:
                S1 = (float(econfig[ii:i1])-1.)/2.
            except ValueError:
                print('Failed to parse S1 in JK coupling, from:\n', config)
                return -1, -1., -1., -1, -1., -1.
            #There should be a value after L1 in <> giving the J1 value
            if ('<' in econfig[i1:i2]) and ('>' in econfig[i1:i2]):
                sJ1 = econfig[econfig.find('<')+1 : econfig.find('>')]
                #J1 may be a fraction and need to be converted
                J1 = _parseFraction(sJ1, errMsg =
                                    'Failed to parse J1 in JK coupling:')
                if np.isnan(J1): return -1, -1., -1., -1, -1., -1.
            break
    
    #If there is no capital L1 for a subgroup, try to find l1 for a single e-
    #(the case with two valence electrons, essentially a j-l case).
    if L1 == -1:
        for i1 in range(i2-1, 0, -1):
            if econfig[i1] in lval:
                L1 = lval.index(econfig[i1])
                S1 = 0.5
                #In this case the J1 value is in the term symbol, before the [K]
                #this should be a fraction since L1 is integer, S1 is 1/2,
                #and J1 = L1 +/- S1.
                sJ1 = termSym.split('[')[0]
                #J1 likely is a fraction
                J1 = _parseFraction(sJ1, errMsg =
                                    'Failed to parse J1 in JK coupling:')
                if np.isnan(J1): return -1, -1., -1., -1, -1., -1.
                break

    if L1 < 0 or L2 < 0 or S1 < 0.0 or S2 < 0.0 or J1 < 0.0 or K < 0.0:
        print('Failed to parse JK coupling, from:\n', config)
    
    return L1, S1, J1, L2, S2, K


def _parseFraction(string, errMsg='failed to parse as a fraction:'):
    """
    Parse a text string fraction and convert it to a floating point number.
    
    This only works for fractions with no whole part in front
    (i.e. a numerator/denominator, or only an integer).
    
    :param string: Text string to convert
    :param errMsg: Error message if this operation fails
    :return: The converted value, will be NaN if the conversion fails.
    """
    val = float('NaN')
    if '/' in string:
        try:
            val = float(string.split('/')[0])/float(string.split('/')[1])
        except ValueError:
            print(errMsg, string)
            return float('NaN')
    else:
        try:
            val = float(string)
        except ValueError:
            print(errMsg, string)
            return float('NaN')
    return val



"""
Converters that read a set of FITS files, in SPIRou APERO spectrum formats,
and save them to text .s formats.
"""

import numpy as np
import astropy.io.fits as fits
from .. import obsSpec as spf

def spirou(flist, flistout=None, ftype=None, nanTreatment='replace',
           removeSegmentSize=200, allowGapSize=3,
           saveFitsHeader=True, writeSpecHeader=False):
    '''
    Convert a .fits, file or list of files, from SPIRou APERO formats into
    text .s files.

    This supports intensity spectrum e.fits files, telluric corrected intensity
    spectrum t.fits files, and polarized spectrum p.fits files.
    This calls the spirou_e, spirou_t, and spirou_p
    functions.

    The .fits files usually contain nan values for pixels where the telluric
    correction failed, or where the spectrum extraction failed.
    This function provides options: to remove those pixels (and optionally 
    remove some of the small fragments of spectrum that get left behind),
    to replace those pixels with placeholder values, or to keep the
    nan values. (Keeping the nan values will cause problems for other
    SpecpolFlow functions!)

    For e.fits and t.fits files: this function estimates uncertainties
    on the spectrum simply as sqrt(N), applies a blaze correction to
    the spectrum, and applies a barycentric velocity correction.

    For p.fits files: this function converts the input polarized spectrum from
    being normalized by the intensity spectrum (e.g. V/I) to being normalized
    by the continuum (e.g. V/Ic, like I/Ic).

    :param flist: .fits file name or list filenames (a string or list of strings)
    :param flistout: optional, an output filename or list of filenames.
                     If this is not provided output names will be generated
                     from the input filenames, replacing .fits with .s
    :param ftype: the type of input .fits file to process.
                  'e' is e.fits files, with intensity spectra, still with
                  separate spectral orders.
                  't' is t.fits files, with telluric corrected spectra,
                  still with separate spectral orders.
                  'p' is p.fits files, with polarized spectra.
                  If this is not provided, the file type will be estimated
                  from the input filename, but the files must end with
                  e.fits, t.fits, or p.fits.
    :param nanTreatment: Flag for how to deal with nan values in the input file,
                         can be: 'remove', 'replace', 'keep'. 
                         nan values can be from places where the spectrum 
                         extraction failed or the telluric correction was too 
                         uncertain.  'remove' deletes pixels with nans.
                         'replace' replaces the nan values with 0, and sets the
                         uncertainties to 100. 
                         'keep' retains the nan values in the spectrum.
    :param removeSegmentSize: If nanTreatment = 'remove', this can leave a lot
                         of little fragments of spectrum in the output, mostly
                         in regions with strong telluric lines.  
                         Fragments that are shorter than removeSegmentSize 
                         pixels will be removed from the output. (Set this to
                         0 to keep all the spectrum)
    :param allowGapSize: If nanTreatment = 'remove', and small fragments
                         of spectrum are also being removed, there can be small
                         gaps of a few pixels in the spectrum. Setting
                         allowGapSize ignores gaps smaller than this value when
                         calculating the size of segments for removeSegmentSize.
    :param saveFitsHeader: Flag for whether to save the .fits header information
                         to a text file, named using the output filename
                         .out (True = save to file).
    :param writeSpecHeader: Flag for whether to write two lines of header in the
                         .s text file (True = include header)
    :return: list of Spectrum objects
    '''
    if isinstance(flist, str): flist = [flist,]
    #Some basic error checking
    if not (isinstance(flist, list) or isinstance(flist, tuple)):
        raise ValueError('in spirou(), flist must be a string '
                         'or list of strings')
    if not flistout is None:
        if isinstance(flistout, str): flistout = [flistout,]
        if not (isinstance(flistout, list) or isinstance(flistout, tuple)):
            raise ValueError('in spirou(), flistout must be a string '
                             'or list of strings with output file names')
        if len(flist) != len(flistout):
            raise ValueError('in spirou(), the list of input file names and '
                             'list of output file names must have the same length')
    
    speclist = []
    for i, fname in enumerate(flist):
        #If an output file name exists use it
        if flistout is None:
            outname = None
        else:
            outname = flistout[i]
        #If no file type is given, try guessing from the file name
        if ftype is None:
            if len(fname) > 6:
                if fname[-6:] == 'p.fits':
                    _ftype = 'p'
                elif fname[-6:] == 'e.fits':
                    _ftype = 'e'
                elif fname[-6:] == 't.fits':
                    _ftype = 't'
        else:
            _ftype = ftype
        
        if _ftype == 'p':
            spec = spirou_p(fname, outname, nanTreatment=nanTreatment,
                            removeSegmentSize=removeSegmentSize,
                            allowGapSize=allowGapSize,
                            saveFitsHeader=saveFitsHeader,
                            writeSpecHeader=writeSpecHeader)
        elif _ftype == 'e':
            spec = spirou_e(fname, outname, nanTreatment=nanTreatment,
                            removeSegmentSize=removeSegmentSize,
                            allowGapSize=allowGapSize,
                            saveFitsHeader=saveFitsHeader,
                            writeSpecHeader=writeSpecHeader)
        elif _ftype == 't':
            spec = spirou_t(fname, outname, nanTreatment=nanTreatment,
                            removeSegmentSize=removeSegmentSize,
                            allowGapSize=allowGapSize,
                            saveFitsHeader=saveFitsHeader,
                            writeSpecHeader=writeSpecHeader)
        else:
            raise ValueError('in spirou(), unrecognized ftype: '
                             '{:} (only can use: p t e)'.format(ftype))

        speclist += [spec]
    return speclist


def spirou_p(fname, outname=None, nanTreatment='replace',
             removeSegmentSize=200, allowGapSize=3,
             saveFitsHeader=True, writeSpecHeader=False):
    '''
    Function to read in a SPIRou p.fits file containing a polarized spectrum,
    save it to a .s file, and return a Spectrum object

    The p.fits files usually contain the polarization spectrum normalized by
    the full Stokes I spectrum, not the continuum of Stokes I
    (e.g. V/I not V/Ic).  This function converts the polarized spectrum to be
    normalized by the continuum (outputting e.g. V/Ic).

    The p.fits files usually contain nan values for pixels where the telluric
    correction failed, or where the spectrum extraction failed.
    This function provides options: to remove those pixels (and optionally 
    remove some of the small fragments of spectrum that get left behind),
    to replace those pixels with placeholder values, or to keep the
    nan values. (Keeping the nan values will cause problems for other
    SpecpolFlow functions!)
    APERO p.fits files also have nan values for wavelengths, so in the 
    'replace' mode wavelengths in nan regions are only estimates, 
    based on the surrounding wavelength values.
    
    :param fname: name of the .fits file to read
    :param outname: name of the file to save the spectrum to.  If not provided
                    the function will default to using the input fname and
                    replacing .fits with .s
    :param nanTreatment: Flag for how to deal with nan values in the input file,
                         can be: 'remove', 'replace', 'keep'. 
                         nan values can be from places where the spectrum 
                         extraction failed or the telluric correction was too 
                         uncertain. 'remove' deletes pixels with nans.
                         'replace' replaces the nan values with 0, and sets the
                         uncertainties to 100. 
                         'keep' retains the nan values in the spectrum.
    :param removeSegmentSize: If nanTreatment = 'remove', this can leave a lot
                         of little fragments of spectrum in the output, mostly
                         in regions with strong telluric lines.  
                         Fragments that are shorter than removeSegmentSize 
                         pixels will be removed from the output. (Set this to
                         0 to keep all the spectrum)
    :param allowGapSize: If nanTreatment = 'remove', and small fragments
                         of spectrum are also being removed, there can be small
                         gaps of a few pixels in the spectrum. Setting
                         allowGapSize ignores gaps smaller than this value when
                         calculating the size of segments for removeSegmentSize.
    :param saveFitsHeader: Flag for whether to save the .fits header information
                         to a text file, named using the output filename
                         .out (True = save to file).
    :param writeSpecHeader: Flag for whether to write two lines of header in the
                         .s text file (True = include header)
    :rtype: Spectrum
    '''
    if isinstance(fname, list) or isinstance(fname, tuple):
        raise  ValueError('spirou_p() can only process one file'
                          ' at a time, not a list of files!')
    elif not isinstance(fname, str):
        raise  ValueError('in spirou_p(), file name must be a string!')
    if outname is None:
        outname = fname+'.s'
        if outname[-7:-2].lower() == '.fits': outname = outname[:-7]+'.s'
    print('converting', fname, 'to', outname)
    if (nanTreatment != 'remove' and nanTreatment != 'replace'
        and nanTreatment != 'keep'):
        raise ValueError('in spirou_p(), unrecognized nanTreatment '
                         'value: {:}'.format(nanTreatment))
    # read the fits file
    fitsSpec = fits.open(fname)
    header = fitsSpec[0].header
    header2 = fitsSpec[1].header
    
    #print('File format info')
    #print(fitsSpec.info())
    #print('Header information')
    #print(repr(header))

    target = header['OBJECT']
    dateUTC = header['DATE-OBS']
    timeUTC = header['UTIME']
    # get the 2D data arrays, with one row for each order
    polar2D    = fitsSpec[1].data
    polarErr2D = fitsSpec[2].data
    specI2D    = fitsSpec[3].data
    specIerr2D = fitsSpec[4].data
    null12D    = fitsSpec[5].data
    null22D    = fitsSpec[6].data
    wl2D       = fitsSpec[7].data
    blaze2D    = fitsSpec[8].data
    fitsSpec.close()

    # trim nans of the start and end of each order
    # (orders are often extracted past the end of the usable flux, and the data
    # are padded with nans so the rows for the orders all have the same length)
    nOrd, nPix = wl2D.shape
    for i in range(nOrd):
        # save this order into a Spectrum object (trimming nans off the edges)
        nuse = np.nonzero(np.isfinite(specI2D[i,:]))[0]
        ord = spf.Spectrum(wl2D[i,:][nuse[0]:nuse[-1]+1],
                           specI2D[i,:][nuse[0]:nuse[-1]+1],
                           polar2D[i,:][nuse[0]:nuse[-1]+1],
                           null12D[i,:][nuse[0]:nuse[-1]+1],
                           null22D[i,:][nuse[0]:nuse[-1]+1],
                           polarErr2D[i,:][nuse[0]:nuse[-1]+1])
        ordBlaze = blaze2D[i,:][nuse[0]:nuse[-1]+1]

        # APERO p.fits files have a continuum normalized I/Ic,
        # but *not* continuum normalized polarization spectra (like V/I, not V/Ic).
        # So convert the polarization spectrum to be continuum normalized,
        # like the intensity spectrum.
        ord.specV = ord.specV*ord.specI
        ord.specN1 = ord.specN1*ord.specI
        ord.specN2 = ord.specN2*ord.specI
        ord.specSig = ord.specSig*ord.specI
        
        # get remaining nan values in the order
        nuse = np.isfinite(ord.wl) & np.isfinite(ord.specI) & np.isfinite(ord.specV)
        nan = np.logical_not(nuse)
        
        if nanTreatment == 'replace':
            # replace NaNs with zeros (for the normalization routine)
            ord.specI[nan]   = 0.0
            ord.specV[nan]   = 0.0
            ord.specN1[nan]  = 0.0
            ord.specN2[nan]  = 0.0
            ord.specSig[nan] = 100.0

            # make up some wavelengths where they don't exist
            # (this version can be a bit slow, assigning one wl value at a time)
            indfix = np.nonzero(np.isnan(ord.wl))[0]
            indok = np.nonzero(np.isfinite(ord.wl))[0]
            for j in indfix:
                lastOk = indok[indok < j][-1]
                nextOk = indok[indok > j][0]
                wlStep = (ord.wl[nextOk] - ord.wl[lastOk])/float(nextOk - lastOk)
                wlGuess = ord.wl[lastOk] + wlStep*(j - lastOk)
                ord.wl[j] = wlGuess

        elif nanTreatment == 'remove':
            # cut the NaNs, and also small fragments of spectrum between nans
            _flag_spec_fragments(nuse, removeSegmentSize, allowGapSize)
            ord = ord[nuse]

        # add this order to the full spectrum
        if i == 0:
            spec = ord
        else:
            spec = spec.concatenate(ord)

    # add a header and save the result
    spec.header = '***Reduced spectrum {:} {:} {:}\n'.format(target, dateUTC, timeUTC)
    spec.save(outname, saveHeader=writeSpecHeader)
    
    if saveFitsHeader:
        outname2 = outname+'.out'
        if outname2[-6:-4] == '.s': outname2 = outname2[:-6]+'.out'
        fout = open(outname2, 'w')
        fout.write('# Main FITS header\n')
        fout.write(repr(header)+'\n')
        fout.write('# Additional spectrum FITS header\n')
        # avoid repeating redundant header information
        for card in header2.cards:
            if (card.keyword not in header) or (card.keyword == 'HISTORY'
                                            or card.keyword == 'COMMENT'):
                fout.write(str(card)+'\n')
        fout.close()
    return spec


def spirou_e(fname, outname=None, nanTreatment='replace',
             removeSegmentSize=100, allowGapSize=3, bervCorr=True,
             saveFitsHeader=True, writeSpecHeader=False):
    '''
    Function to read in a SPIRou e.fits file containing an intensity spectrum,
    save it to a .s file, and return a Spectrum object

    This function estimates uncertainties on the spectrum simply as the 
    square root of the flux, then applies a blaze correction to the spectrum.
    
    The e.fits files usually contain nan values for pixels where the spectrum
    extraction failed (most commonly near order edges where the flux is low).
    This function provides options: to remove those pixels (and optionally
    remove some of the small fragments of spectrum that get left behind), 
    to replace those pixels with placeholder values, or to keep the
    nan values. (Keeping the nan values will cause problems for other
    SpecpolFlow functions!)

    :param fname: name of the .fits file to read
    :param outname: name of the file to save the spectrum to.  If not provided
                    the function will default to using the input fname and
                    replacing .fits with .s
    :param nanTreatment: Flag for how to deal with nan values in the input file,
                         can be: 'remove', 'replace', 'keep'. 
                         'remove' deletes pixels with any nans.
                         'replace' replaces the nan values with 0, and sets the
                         uncertainties to 100. 
                         'keep' retains the nan values in the spectrum.
    :param removeSegmentSize: If nanTreatment='remove', this can leave
                         little fragments of spectrum in the output.
                         Fragments that are shorter than removeSegmentSize
                         pixels will be removed from the output. (Set this to
                         0 to keep all the spectrum)
    :param allowGapSize: If nanTreatment = 'remove', and small fragments
                         of spectrum are also being removed, there can be small
                         gaps of a few pixels in the spectrum. Setting
                         allowGapSize ignores gaps smaller than this value when
                         calculating the size of segments for removeSegmentSize.
    :param bervCorr:     Flag for whether to apply a barycentric radial velocity
                         correction to the spectrum (using the BERV from the
                         fits header, True = apply correction).  Without this
                         correction the observation is in the observer's rest
                         frame, rather than the solar system barycentric rest
                         frame.
    :param saveFitsHeader: Flag for whether to save the .fits header information
                         to a text file, named using the output filename
                         .out (True = save to file).
    :param writeSpecHeader: Flag for whether to write two lines of header in the
                         .s text file (True = include header)
    :rtype: Spectrum
    '''
    if isinstance(fname, list) or isinstance(fname, tuple):
        raise  ValueError('spirou_e() can only process one file'
                          ' at a time, not a list of files!')
    elif not isinstance(fname, str):
        raise  ValueError('in spirou_e(), file name must be a string!')
    if outname is None:
        outname = fname+'.s'
        if outname[-7:-2].lower() == '.fits': outname = outname[:-7]+'.s'
    print('converting', fname, 'to', outname)
    if (nanTreatment != 'remove' and nanTreatment != 'replace'
        and nanTreatment != 'keep'):
        raise ValueError('in spirou_e(), unrecognized nanTreatment '
                         'value: {:}'.format(nanTreatment))
    # read the fits file
    fitsSpec = fits.open(fname)
    header = fitsSpec[0].header
    header2 = fitsSpec[1].header
    header3 = fitsSpec[2].header
    
    #print('File format info')
    #print(fitsSpec.info())
    #print('Header information')
    #print(repr(header))

    target = header['OBJECT']
    dateUTC = header['DATE-OBS']
    timeUTC = header['UTIME']
    # get the 2D data arrays, with one row for each order
    specI2D = fitsSpec[1].data
    wl2D    = fitsSpec[5].data
    blaze2D = fitsSpec[9].data
    fitsSpec.close()
    
    # trim nans of the start and end of each order
    # (orders are often extracted past the end of the usable flux, and the data
    # are padded with nans so the rows for the orders all have the same length)
    nOrd, nPix = wl2D.shape
    for i in range(nOrd):
        # save this order into a Spectrum object (trimming nans off the edges)
        nuse = np.nonzero(np.isfinite(specI2D[i,:]))[0]
        ord = spf.Spectrum(wl2D[i,:][nuse[0]:nuse[-1]+1],
                           specI2D[i,:][nuse[0]:nuse[-1]+1],
                           np.zeros(nuse[-1] - nuse[0] + 1),
                           np.zeros(nuse[-1] - nuse[0] + 1),
                           np.zeros(nuse[-1] - nuse[0] + 1),
                           100.*np.ones(nuse[-1] - nuse[0] + 1))
        ordBlaze = blaze2D[i,:][nuse[0]:nuse[-1]+1]

        # get remaining nan values in the order
        nuse = np.isfinite(ord.wl) & np.isfinite(ord.specI) & np.isfinite(ordBlaze)
        nan = np.logical_not(nuse)
        
        # estimate uncertainties
        iok = (nuse) & (ord.specI > 0.0)
        ord.specSig[iok] = np.sqrt(ord.specI[iok])
        
        # apply blaze correction
        ord.specI = ord.specI/ordBlaze
        ord.specSig = ord.specSig/ordBlaze
        # and set uncertainty estimates for pixels with negative flux to be large
        #ord.specSig[np.logical_not(iok)] = np.percentile(ord.specSig[iok], 99.9)
        ord.specSig[np.logical_not(iok)] = 0.2*np.percentile(ord.specI[iok], 99.0)

        if nanTreatment == 'replace':
            # replace NaNs with zeros (for the normalization routine)
            ord.specI[nan] = 0.0
            ord.specSig[nan] = 100.0
        
        elif nanTreatment == 'remove':
            # cut the NaNs, and also small fragments of spectrum between nans
            _flag_spec_fragments(nuse, removeSegmentSize, allowGapSize)
            ord = ord[nuse]

        # add this order to the full spectrum
        if i == 0:
            spec = ord
        else:
            spec = spec.concatenate(ord)

    # The e.fits spectra default to being in the observer's rest frame rather
    # than the solar system barycentric rest frame. Optionally, correct for this
    # velocity shift (p.fits files seem to already have this correction).
    if bervCorr:
        try:
            berv = header2['BERV']
            bjd = header2['BJD']
            print("applying barycentric radial velocity correction "
                  "{:8.4f} km/s, Barycentric Julian date {:.6f}".format(berv, bjd))
        except:
            berv = 0.0
            print("Warning: Failed to find barycentric radial velocity, "
                  "no correction applied!")
        c = 299792.458
        spec.wl = spec.wl + spec.wl*berv/c

    # add a header and save the result
    spec.header = '***Reduced spectrum {:} {:} {:}\n'.format(target, dateUTC, timeUTC)
    spec.save(outname, saveHeader=writeSpecHeader)

    if saveFitsHeader:
        outname2 = outname+'.out'
        if outname2[-6:-4] == '.s': outname2 = outname2[:-6]+'.out'
        fout = open(outname2, 'w')
        fout.write('# Main FITS header\n')
        fout.write(repr(header)+'\n')
        fout.write('# Additional spectrum FITS header\n')
        # avoid repeating redundant header information
        for card in header2.cards:
            if (card.keyword not in header) or (card.keyword == 'HISTORY'
                                            or card.keyword == 'COMMENT'):
                fout.write(str(card)+'\n')
        fout.close()
    return spec


def spirou_t(fname, outname=None, nanTreatment='replace',
             removeSegmentSize=200, allowGapSize=3, bervCorr=True,
             saveFitsHeader=True, writeSpecHeader=False):
    '''
    Function to read in a SPIRou t.fits file containing a telluric corrected
    intensity spectrum, save it to a .s file, and return a Spectrum object

    This function estimates uncertainties on the spectrum simply as the 
    square root of the flux, then applies a blaze correction to the spectrum.
    
    The t.fits files usually contain the telluric corrected spectrum
    and a copy of the telluric line spectrum removed from the original
    observation.  This function saves both to two different files
    (with the telluric spectrum in [outname].telluric).

    The t.fits files usually contain nan values for pixels where the telluric
    correction failed, or where the spectrum extraction failed.
    This function provides options: to remove those pixels (and optionally
    remove some of the small fragments of spectrum that get left behind),
    to replace those pixels with placeholder values, or to keep the
    nan values. (Keeping the nan values will cause problems for other
    SpecpolFlow functions!)
    
    :param fname: name of the .fits file to read
    :param outname: name of the file to save the spectrum to.  If not provided
                    the function will default to using the input fname and
                    replacing .fits with .s
    :param nanTreatment: Flag for how to deal with nan values in the input file,
                         can be: 'remove', 'replace', 'keep'. 
                         nan values can be from bad pixels or from places where
                         the telluric correction was too uncertain.
                         'remove' deletes pixels with any nans.
                         'replace' replaces the nan values with 0, and sets the
                         uncertainties to 100. 
                         'keep' retains the nan values in the spectrum.
    :param removeSegmentSize: If nanTreatment = 'remove', this can leave a lot
                         of little fragments of spectrum in the output, mostly
                         in regions with strong telluric lines.  
                         Fragments that are shorter than removeSegmentSize 
                         pixels will be removed from the output. (Set this to
                         0 to keep all the spectrum)
    :param allowGapSize: If nanTreatment = 'remove', and small fragments
                         of spectrum are also being removed, there can be small
                         gaps of a few pixels in the spectrum. Setting
                         allowGapSize ignores gaps smaller than this value when
                         calculating the size of segments for removeSegmentSize.
    :param bervCorr:     Flag for whether to apply a barycentric radial velocity
                         correction to the spectrum (using the BERV from the
                         fits header, True = apply correction).  Without this
                         correction the observation is in the observer's rest
                         frame, rather than the solar system barycentric rest
                         frame.
    :param saveFitsHeader: Flag for whether to save the .fits header information
                         to a text file, named using the output filename
                         .out (True = save to file).
    :param writeSpecHeader: Flag for whether to write two lines of header in the
                         .s text file (True = include header)
    :rtype: Spectrum
    '''
    if isinstance(fname, list) or isinstance(fname, tuple):
        raise  ValueError('spirou_t() can only process one file'
                          ' at a time, not a list of files!')
    elif not isinstance(fname, str):
        raise  ValueError('in spirou_t(), file name must be a string!')
    if outname is None:
        outname = fname+'.s'
        if outname[-7:-2].lower() == '.fits': outname = outname[:-7]+'.s'
    print('converting', fname, 'to', outname, 'and', outname+'.telluric')
    if (nanTreatment != 'remove' and nanTreatment != 'replace'
        and nanTreatment != 'keep'):
        raise ValueError('in spirou_t(), unrecognized nanTreatment '
                         'value: {:}'.format(nanTreatment))
    # read the fits file
    fitsSpec = fits.open(fname)
    header = fitsSpec[0].header
    header2 = fitsSpec[1].header
    
    #print('File format info')
    #print(fitsSpec.info())
    #print('Header information')
    #print(repr(header))

    target = header['OBJECT']
    dateUTC = header['DATE-OBS']
    timeUTC = header['UTIME']
    # get the 2D data arrays, with one row for each order
    specI2D = fitsSpec[1].data
    wl2D    = fitsSpec[2].data
    blaze2D = fitsSpec[3].data
    tellurc2D = fitsSpec[4].data
    # This extra OH emission spectrum could be useful,
    # but I'm not sure how to interpret or use it.
    OHline2D = fitsSpec[5].data
    fitsSpec.close()
    
    # trim nans of the start and end of each order
    # (orders are often extracted past the end of the usable flux, and the data
    # are padded with nans so the rows for the orders all have the same length)
    nOrd, nPix = wl2D.shape
    for i in range(nOrd):
        # save this order into a Spectrum object (trimming nans off the edges)
        nuse = np.nonzero(np.isfinite(specI2D[i,:]))[0]
        ord = spf.Spectrum(wl2D[i,:][nuse[0]:nuse[-1]+1],
                           specI2D[i,:][nuse[0]:nuse[-1]+1],
                           np.zeros(nuse[-1] - nuse[0] + 1),
                           np.zeros(nuse[-1] - nuse[0] + 1),
                           np.zeros(nuse[-1] - nuse[0] + 1),
                           100.*np.ones(nuse[-1] - nuse[0] + 1))
        ordTell = spf.Spectrum(wl2D[i,:][nuse[0]:nuse[-1]+1],
                               tellurc2D[i,:][nuse[0]:nuse[-1]+1],
                               np.zeros(nuse[-1] - nuse[0] + 1),
                               np.zeros(nuse[-1] - nuse[0] + 1),
                               np.zeros(nuse[-1] - nuse[0] + 1),
                               np.zeros(nuse[-1] - nuse[0] + 1))
        ordBlaze = blaze2D[i,:][nuse[0]:nuse[-1]+1]

        # get remaining nan values in the order
        nuse = np.isfinite(ord.wl) & np.isfinite(ord.specI) & np.isfinite(ordBlaze)
        nan = np.logical_not(nuse)
        
        # estimate uncertainties
        iok = (nuse) & (ord.specI > 0.0)
        ord.specSig[iok] = np.sqrt(ord.specI[iok])
        
        # apply blaze correction
        ord.specI = ord.specI/ordBlaze
        ord.specSig = ord.specSig/ordBlaze
        # and set uncertainty estimates for pixels with negative flux to be large
        #ord.specSig[np.logical_not(iok)] = np.percentile(ord.specSig[iok], 99.9)
        ord.specSig[np.logical_not(iok)] = 0.2*np.percentile(ord.specI[iok], 99.0)

        if nanTreatment == 'replace':
            # replace NaNs with zeros (for the normalization routine)
            ord.specI[nan] = 0.0
            ord.specSig[nan] = 100.0
            ordTell.specI[nan] = 0.0
        
        elif nanTreatment == 'remove':
            # cut the NaNs, and also small fragments of spectrum between nans
            _flag_spec_fragments(nuse, removeSegmentSize, allowGapSize)
            ord = ord[nuse]
            ordTell = ordTell[nuse]

        # add this order to the full spectrum
        if i == 0:
            spec = ord
            specTell = ordTell
        else:
            spec = spec.concatenate(ord)
            specTell = specTell.concatenate(ordTell)

    # The t.fits spectra default to being in the observer's rest frame rather
    # than the solar system barycentric rest frame. Optionally, correct for this
    # velocity shift (p.fits files seem to already have this correction).
    if bervCorr:
        try:
            berv = header2['BERV']
            bjd = header2['BJD']
            print("applying barycentric radial velocity correction "
                  "{:8.4f} km/s, Barycentric Julian date {:.6f}".format(berv, bjd))
        except:
            berv = 0.0
            print("Warning: Failed to find barycentric radial velocity, "
                  "no correction applied!")
        c = 299792.458
        spec.wl = spec.wl + spec.wl*berv/c
        specTell.wl = specTell.wl + specTell.wl*berv/c

    # add a header and save the result
    spec.header = '***telluric corrected spectrum {:} {:} {:}\n'.format(target, dateUTC, timeUTC)
    spec.save(outname, saveHeader=writeSpecHeader)
    specTell.header = '***telluric spectrum for {:} {:} {:}\n'.format(target, dateUTC, timeUTC)
    specTell.save(outname+'.telluric', saveHeader=writeSpecHeader)

    if saveFitsHeader:
        outname2 = outname+'.out'
        if outname2[-6:-4] == '.s': outname2 = outname2[:-6]+'.out'
        fout = open(outname2, 'w')
        fout.write('# Main FITS header\n')
        fout.write(repr(header)+'\n')
        fout.write('# Additional spectrum FITS header\n')
        # avoid repeating redundant header information
        for card in header2.cards:
            if (card.keyword not in header) or (card.keyword == 'HISTORY'
                                            or card.keyword == 'COMMENT'):
                fout.write(str(card)+'\n')
        fout.close()

    return spec


def _flag_spec_fragments(nuse, removeSegmentSize, allowGapSize):
    '''
    Helper function for converters.  Takes an array of bool flags for whether
    pixels are used.  Look for small fragments of spectrum that are used,
    and flag them to not be used.  Modifies the array of flags in place.
    '''
    istartOk = 0
    sizeOk = 0
    sizeNaN = 0
    # go through each pixel and flag short fragments of spectrum to remove.
    for i in range(nuse.size):
        if nuse[i] == False:
            sizeNaN += 1
            # If we reach the end of an ok region, but that region is too
            # small, don't use it.
            if sizeOk > 0 and sizeOk < removeSegmentSize:
                # But wait until we have found a large enough gap
                # before removing the small ok region. (if the NaN gap
                # is too small, keep adding to the current ok region)
                if sizeNaN > allowGapSize:
                    nuse[istartOk:i] = False
            # If we have found too large a NaN gap, reset the ok region
            if sizeNaN > allowGapSize and sizeOk > 0:
                sizeOk = 0
        else:
            if sizeOk == 0: istartOk = i
            sizeOk += 1
            sizeNaN = 0
    # If we haven't closed the last region to remove
    if sizeNaN > 0 and sizeOk > 0:
        nuse[istartOk:i+1] = False
    # If we have a fragment as the last region
    elif sizeOk > 0 and  sizeOk < removeSegmentSize:
        nuse[istartOk:i+1] = False        
    return



#!/usr/bin/python3
#
#A wrapper for quickly plotting a set of LSD profiles from the command line

import argparse
import matplotlib.pyplot as plt
try:
    import specpolFlow as pol
except ModuleNotFoundError:
    #If specpolFlow is not installed or in the python path
    #try guessing the location and adding it temporarily
    import sys, os
    loc0 = os.path.dirname(__file__) 
    sys.path.insert(0, os.path.join(loc0, '..', '..'))
    import specpolFlow as pol

def plot_lsd_cli():
    """
    Function for quickly reading and plotting LSD profiles from a terminal.
    """
    #Take input file names as command line arguments,
    #with some additional optional control parameters.
    parser = argparse.ArgumentParser(description='Plot a set of LSD profiles.')
    parser.add_argument("fileList", nargs='*',
                        help="LSD profile files to plot, can be more than one file.")
    parser.add_argument("-l", "--legend", action='store_true',
                        help="Optionally, show a legend of file names")
    parser.add_argument("-s", "--save", default=None,
                        help="Optionally, save the plot to this file")
    args = parser.parse_args()
    #Process the command line parameters
    fileList = []
    for fileName in args.fileList:
        fileList += [fileName]
    showLegend = args.legend
    saveName = args.save

    fig = None
    axs = None
    for fileName in fileList:
        lsd = pol.read_lsd(fileName)
        fig, axs = lsd.plot(fig=fig, ls='-', label=fileName)

    if showLegend:
        axs[-1].legend(loc='lower left')
    if saveName is not None:
        print('saving to', saveName)
        fig.savefig(saveName)
    
    plt.show()
    return

#For running this Python script as a terminal program
if __name__ == "__main__":
    plot_lsd_cli()



"""
Read a set of FITS files, in the ESPaDOnS spectra format from the CADC
Can be modified for other spectra formats,
if you know the structure of the records in the FITS file.
"""

import numpy as np
import astropy.io.fits as fits
from .. import obsSpec as spf

def espadons(flist, flistout=None, writeSpecHeader=False):
    """
    Convert a list of .fits files in the CADC ESPaDOnS format into
    text .s files. 

    The code provides two files in .s format:  
     
    * The UPENA normalized spectrum (n.s), with automated radial velocity corrections from the telluric lines.
    * The UPENA unnormalized spectrum (u.s), using the automated radial velocity correction from the normalized spectrum.
    This is done starting from the unnormalized spectrum without the automated 
    radial velocity correction, to which we apply the radial velocity
    correction determined from the normalized spectrum. The reason behind this
    is that the UPENA automated radial velocity determination performed on 
    unnormalized spectra does not produce consistently reliable results. 
    The content of the fits header is also saved in a .out text file.
    If flistout=None (default), the files are written at the same path as
    the fits-format data, with the '.fits' stripped, 
    and 'n.s' and 'u.s' appended to the filename root. 
    If flistout is a list of paths ending with a rootname, the files will be
    saved at that path with that rootname, and 'n.s' and 'u.s' appended to
    the filename root.

    :param flist: (list of strings) a list of ESPaDOnS filenames
    :param flistout: (list of strings) optional, list of output file rootnames
    :param writeSpecHeader: Flag for whether to write two lines of header in the
                         .s text file (True = include header)
    """

    if isinstance(flist, str): flist = [flist,]
    #Some basic error checking
    if not isinstance(flist, list):
        raise ValueError('in espadons(), the flist argument must be a Python '
                         'list of input file names (or a single file name)')
    if not (flistout is None):
        if isinstance(flistout, str): flistout = [flistout,]
        if not isinstance(flistout, list):
            raise ValueError('in espadons(), the flistout argument must be a '
                             'Python list of output file names')
        if len(flist) != len(flistout):
            raise ValueError('in espadons(), the list of input file names and '
                             'list of output file names must have the same length')
    
    for i, fname in enumerate(flist):
        print('converting ', fname.strip())
        if flistout is None:
            # striping of white spaces
            fnameOut = fname.strip()
            # removing the '.fits' from the end of the string, 
            # to create the root name for the output files. 
            fnameOut = fnameOut.rstrip('.fits')
        else:
            fnameOut = flistout[i]
        # open the fits file with astropy
        fitsSpec = fits.open(fname.strip())

        header = fitsSpec[0].header

        # Useful for debugging
        #print('File format info')
        #print(fitsSpec.info())
        #print('Header information')
        #print(repr(header))

        # We need to extract the radial velocity correction 
        # that was determined from the normalized spectrum,
        # so that we can apply it to the unnormalized spectrum

        # the radial velocity correction is in the comment section
        # of the header. Using the 'COMMENT' keyword returns a 
        # list of strings
        comments = header['COMMENT']
        # searching for the two strings in the comments that contains
        # the radial velocity correction:
        matches = [match for match in comments if
                   "radial velocity correction from telluric lines" in match]
        # the one calculated for the normalized spectrum is the first instance
        radvel = float(matches[0].split(':')[1])
        
        target = header['OBJECT']
        dateUTC = header['DATE-OB1']
        timeUTC = header['UTIME1']
        # extracting the table of data (24 columns)
        specTab = fitsSpec[0].data

        # The normalized spectrum with radial velocity correction from telluric lines
        # is the first 6 columns
        wln = specTab[0]
        specIn = specTab[1]
        specVn = specTab[2]
        specN1n = specTab[3]
        specN2n = specTab[4]
        specSign = specTab[5]
    
        # The unnormalized spectrum *without* the radial velocity correction
        # from the telluric lines is the last of 4 data blocks
        # (Use this since the radial velocity correction for the unnormalized 
        #  spectrum is unreliable)
        wlu = specTab[18]
        specIu = specTab[19]
        specVu = specTab[20]
        specN1u = specTab[21]
        specN2u = specTab[22]
        specSigu = specTab[23]
    
        fitsSpec.close()

        # Now we apply the velocity correction to the unnormalized data
        c = 299792.458  #speed of light in km/s, since radvel is in km/s
        wlu = wlu + wlu*radvel/c

        spec_n = spf.Spectrum(wln, specIn, specVn, specN1n, specN2n, specSign)
        spec_n.header = '***Reduced spectrum {:} {:} {:}\n'.format(
            target, dateUTC, timeUTC)
        spec_n.save(fnameOut+'n.s', saveHeader=writeSpecHeader)

        spec_u = spf.Spectrum(wlu, specIu, specVu, specN1u, specN2u, specSigu)
        spec_u.header = '***Reduced spectrum {:} {:} {:}\n'.format(
            target, dateUTC, timeUTC)
        spec_u.save(fnameOut+'u.s', saveHeader=writeSpecHeader)
        
        outHeader = open(fnameOut+'.out','w')
        outHeader.write(repr(header))
        outHeader.close()
    return



#!/usr/bin/python3
#
# A simple tool for measuring radial velocities from LSD profiles.

import argparse
try:
    import specpolFlow as pol
except ModuleNotFoundError:
    #If specpolFlow is not installed or in the python path
    #try guessing the location and adding it temporarily
    import sys, os
    loc0 = os.path.dirname(__file__) 
    sys.path.insert(0, os.path.join(loc0, '..', '..'))
    import specpolFlow as pol

def rv_fit_cli():
    """Main function for running LSD.fit_gaussian_rv() as a terminal program

    Takes no arguments, but uses command line parameters instead.
    """
    #Take input file names as command line arguments,
    #with some additional optional control parameters.
    parser = argparse.ArgumentParser(description='Fit a Gaussian to an LSD line profile, to determine radial velocity.  Prints the radial velocity in the same velocity units as the LSD profile.')
    parser.add_argument("fileList", nargs='*',
                        help='LSD profile files, to determine RVs for.  Can be more than one file.')
    parser.add_argument("-v", "--velRange", nargs=2, type=float, metavar=('VEL1', 'VEL2'),
                        help='Optional starting and ending velocity for the range used in the fit')
    parser.add_argument("-p", "--plotFit", action='store_true',
                        help='Optional plot the fit to the LSD profile')
    args = parser.parse_args()
    #Process the command line parameters
    fileList = []
    for fileName in args.fileList:
        fileList += [fileName]
    velRange = args.velRange
    plotFit = args.plotFit

    print('fileName                               RV    error')
    #Loop over the files provided, and fit a radial velocity for each
    for fileName in fileList:
        lsd = pol.read_lsd(fileName)
        
        #Run the radial velocity fitting routine
        fitVel, fitVelErr = lsd.fit_gaussian_rv(velRange, plotFit=plotFit)

        print('{:30} {:10.4f} {:8.4f}'.format(fileName, fitVel, fitVelErr))
    return

#For running as a terminal program
if __name__ == "__main__":
    rv_fit_cli()



"""
An interactive program for cleaning problem lines from a LSD line mask
and tweaking (fitting) depths of some lines.  This opens its main
program in another window.  The LSDpy package must be installed
(in your Python path) to use this program.  
"""

import numpy as np
from matplotlib.figure import Figure

from . import cleanMaskUISubroutines as cleanSub
from . import profileLSD
from . import mask as maskTools
from . import obsSpec

###################################
###################################

def cleanMaskUI(maskName, obsName, outMaskName=None, inExcludeName=None,
                outExcludeName='excludeRanges.dat', batchMode=False):
    """
    Run an interactive tool for LSD line mask cleaning.

    This opens a window with a plot of the line mask, a reference observation,
    and a model LSD spectrum (convolution of a line mask and LSD profile).
    The parameters for the LSD calculation are taken from the defaults in
    the cleanMaskUISubroutines lsdParams() function.

    This saves a copy of the cleaned mask to a file, and also returns
    a copy as a Mask object and the regions excluded from the mask
    as an ExcludeMaskRegions object.

    :param maskName: File name, or a Mask object, for the input line mask
                     to clean.
    :param obsName: File name for the reference observation to compare with.
    :param outMaskName: File name for the output cleaned mask,
                        defaults to [maskName].clean
    :param inExcludeName: File name for reading a set of regions to be excluded
                          from the line mask.
    :param outExcludeName: File name for saving the set of regions excluded 
                           from the mask. (may be the same as inExcludeName)
    :param batchMode: Flag for skipping the GUI. False = open the GUI (the default).
                      True = skip the GUI, just read inExcludeName and apply 
                      those exclude regions.
    :return: a Mask object with the cleaned mask,
             and an ExcludeMaskRegions with the selected exclude regions.
    """
    
    if outMaskName is None:
        if isinstance(maskName, maskTools.Mask):
            outMaskName = 'mask.clean'
        else:
            outMaskName = maskName+'.clean'
    
    #Read the input mask
    if isinstance(maskName, maskTools.Mask):
        mask = maskName
    else:
        mask = maskTools.read_mask(maskName)
    
    #Read a list of wavelength regions to exclude from the mask 
    #(if the file exists), or generate one from default values.
    try:
        if inExcludeName is None: raise OSError #fall back to except block
        regions = maskTools.read_exclude_mask_regions(inExcludeName)
    except OSError:
        regionsBalmer = maskTools.get_Balmer_regions_default(1000.)
        regionsTelluric = maskTools.get_telluric_regions_default()
        regions = regionsBalmer + regionsTelluric
    #Update the mask with these initial exclude regions
    mask = mask.clean(regions)
    #Make a list of lists from the exclude regions, used internally.
    #The inner list should contain only 2 entries: a start and end wavelength.
    #(It could be cleaner to use ExcludeMaskRegions internally, but one would
    #have to be careful to modify the object, not replace it with a copy)
    excludeRanges = []
    for region in regions:
        excludeRanges += [[region.start, region.stop]]
    #Sort the exclude regions by starting wavelength (to be safe)
    excludeRanges = sorted(excludeRanges, key=lambda ran: ran[0])
    
    if batchMode == False:
        #Read the comparison observation
        obs = obsSpec.read_spectrum(obsName)
        
        #Make an initial LSD model spectrum
        print('Generating test LSD profile...')
        lsdp = cleanSub.lsdParams(obsName, outMaskName)
        #first save the current updated mask, so we can run LSDpy on it
        mask.save(outMaskName)
        #then run LSDpy, with specific 'default' parameters
        lsdProf, modelSpec = profileLSD.run_lsdpy(obs=lsdp.obs, mask=lsdp.mask,
            outLSDName=lsdp.outName, velStart=lsdp.velStart, velEnd=lsdp.velEnd,
            velPixel=lsdp.velPixel, normDepth=lsdp.normDepth,
            normLande=lsdp.normLande, normWave=lsdp.normWave,
            removeContPol=lsdp.removeContPol, trimMask=lsdp.trimMask,
            sigmaClipIter=lsdp.sigmaClipIter, sigmaClip=lsdp.sigmaClip,
            outModelName=lsdp.outModelName,
            plotLSD=lsdp.plotLSD, outPlotLSDName=lsdp.outPlotLSDName)
        
        #Make a plot of the reference observation, mask, and LSD model spectrum
        fig =  Figure()
        ax1 = fig.add_subplot(111)
        pltObsI = ax1.plot(obs.wl, obs.specI, 'k')
        maskUsed = mask[mask.iuse == 1]
        maskNot = mask[mask.iuse == 0]
        pltMaskN = ax1.vlines(maskNot.wl, ymin=1.0-maskNot.depth, ymax=1.0,
                              colors='r', lw=1)
        pltMaskU = ax1.vlines(maskUsed.wl, ymin=1.0-maskUsed.depth, ymax=1.0,
                              colors='b', lw=1)
        pltMaskF = ax1.vlines([], ymin=[], ymax=1.0,
                              colors='aqua', lw=1)
        pltModelI = ax1.plot(modelSpec.wl,modelSpec.specI, 'm', lw=1)
        ax1.set_xlabel('Wavelength')
        ax1.set_ylabel('Flux')

        #initialize an array of flags for fitting LSD line depths
        fitDepthFlags = np.zeros_like(mask.wl, dtype=int)
        
        #Run the main interactive program
        cleanSub.makeWin(fig, ax1, mask, obs, lsdp, lsdProf, pltMaskU,
                         pltMaskN, pltMaskF, pltModelI, excludeRanges,
                         outExcludeName, fitDepthFlags)

    #make an ExcludeMaskRegions object to return
    wlStarts = np.array([ran[0] for ran in excludeRanges])
    wlEnds = np.array([ran[1] for ran in excludeRanges])
    labels = np.array(['cleanMaskUI']*len(excludeRanges),dtype=object)
    regions = maskTools.ExcludeMaskRegions(wlStarts, wlEnds, labels)

    #Save the final result
    mask.save(outMaskName)
    return mask, regions



#!/usr/bin/python3
#
# Convert SPIRou .fits spectrum files from the APERO format to a text .s format.

import argparse
import matplotlib.pyplot as plt
try:
    import specpolFlow as pol
except ModuleNotFoundError:
    #If specpolFlow is not installed or in the python path
    #try guessing the location and adding it temporarily
    import sys, os
    loc0 = os.path.dirname(__file__) 
    sys.path.insert(0, os.path.join(loc0, '..', '..'))
    import specpolFlow as pol


def spirou_cli():
    """
    Main function for running the SPIRou converters as a terminal program.
    
    Takes no arguments, but uses command line parameters instead.
    """
    import argparse
    parser = argparse.ArgumentParser(description='Convert FITS file spectra '
                'from the SPIRou APERO format to text .s files. Output files '
                'as [filename].s and [filename].out for header information. \n'
                'Supports e.fits (extracted spectrum in separate orders), '
                't.fits (telluric corrected spectrum in separate orders), '
                'and p.fits (polarized spectrum) files. \n'
                'The FITS files contain nan values where the spectrum '
                'extraction or telluric correction failed.  There are options '
                'to either remove or replace those nan values.')
    parser.add_argument("observation", nargs='*',
                        help='file or list of FITS files to process.')
    parser.add_argument('-t', '--type', default=None,
                        help='optional, specify the type of the fits file to '
                        'process.  Can be e, t, or p.  If not specified the '
                        'type will be inferred from the file name '
                        '(if the name ends in e.fits, t.fits or p.fits)')
    parser.add_argument('-n', '--nan', metavar='TREATMENT', default='replace',
                        help='specify how to treat nan values, can be '
                        '"replace" (replace nan with 0 and uncertainties '
                        'with 100), "remove" (delete pixels with nan and '
                        'some small fragments), or "keep".')
    parser.add_argument('-s', '--removeSegment', metavar='NUMPIX',
                        type=int, default=200,
                        help='optional, only used with "-n remove". Removing '
                        'nan values can leave small fragments of spectrum, '
                        'which are hard to normalize or analyze. '
                        'Fragments smaller than this number of pixels will '
                        'also be removed. Set this to 0 to include all pixels '
                        '(except nans).')
    parser.add_argument('-g', '--allowGap', metavar='NUMPIX',
                        type=int, default=3,
                        help='optional, only used with "-n remove". Removing '
                        'nan values can leave gaps of a few pixels in the '
                        'spectrum.  Gaps smaller than this will be ignored '
                        'when calculating the size of fragments to remove '
                        'from the spectrum.')
    parser.add_argument('-o', '--outputHeader', action='store_true',
                        help='save the header information from the fits file '
                        'to a text file (with a name ending in .out).')
    args = parser.parse_args()

    flist = args.observation
    ftype = args.type
    if isinstance(ftype, str): ftype = ftype.lower()
    nanTreatment = args.nan.lower()
    removeSegmentSize = args.removeSegment
    allowGapSize = args.allowGap
    saveFitsHeader = args.outputHeader

    if flist == []:
        print('No files given')
    else:
        #Run the conversion function
        pol.converters.spirou(flist, ftype=ftype, nanTreatment=nanTreatment,
                              removeSegmentSize=removeSegmentSize,
                              allowGapSize=allowGapSize,
                              saveFitsHeader=saveFitsHeader)
    return

################

#For running this Python script as a terminal program
if __name__ == "__main__":
    spirou_cli()
